from ceph_deploy.util import templates


def install(distro, version_kind, version, adjust_repos, **kw):
    packages = kw.get('components', [])
    distro.packager.clean()
    distro.packager.install(packages)


def mirror_install(distro, repo_url,
                   gpg_url, adjust_repos, extra_installs=True, **kw):
    packages = kw.get('components', [])
    repo_url = repo_url.strip('/')  # Remove trailing slashes
    gpgcheck = kw.pop('gpgcheck', 1)

    distro.packager.clean()

    if adjust_repos:
        distro.packager.add_repo_gpg_key(gpg_url)

        ceph_repo_content = templates.ceph_repo.format(
            repo_url=repo_url,
            gpg_url=gpg_url,
            gpgcheck=gpgcheck,
        )

        distro.conn.remote_module.write_yum_repo(ceph_repo_content)

    if extra_installs and packages:
        distro.packager.install(packages)


def repo_install(distro, reponame, baseurl, gpgkey, **kw):
    # do we have specific components to install?
    # removed them from `kw` so that we don't mess with other defaults
    packages = kw.pop('components', [])

    # Get some defaults
    name = kw.pop('name', '%s repo' % reponame)
    enabled = kw.pop('enabled', 1)
    gpgcheck = kw.pop('gpgcheck', 1)
    install_ceph = kw.pop('install_ceph', False)
    proxy = kw.pop('proxy', '') # will get ignored if empty
    _type = 'repo-md'
    baseurl = baseurl.strip('/')  # Remove trailing slashes

    distro.packager.clean()

    if gpgkey:
        distro.packager.add_repo_gpg_key(gpgkey)

    repo_content = templates.custom_repo(
        reponame=reponame,
        name=name,
        baseurl=baseurl,
        enabled=enabled,
        gpgcheck=gpgcheck,
        _type=_type,
        gpgkey=gpgkey,
        proxy=proxy,
        **kw
    )

    distro.conn.remote_module.write_yum_repo(
        repo_content,
        "%s.repo" % reponame
    )

    # Some custom repos do not need to install ceph
    if install_ceph and packages:
        distro.packager.install(packages)



import logging
from ceph_deploy import exc
from ceph_deploy import conf
from ceph_deploy.cliutil import priority
from ceph_deploy import hosts

LOG = logging.getLogger(__name__)


def admin(args):
    conf_data = conf.ceph.load_raw(args)

    try:
        with open('%s.client.admin.keyring' % args.cluster, 'rb') as f:
            keyring = f.read()
    except:
        raise RuntimeError('%s.client.admin.keyring not found' %
                           args.cluster)

    errors = 0
    for hostname in args.client:
        LOG.debug('Pushing admin keys and conf to %s', hostname)
        try:
            distro = hosts.get(hostname, username=args.username)

            distro.conn.remote_module.write_conf(
                args.cluster,
                conf_data,
                args.overwrite_conf,
            )

            distro.conn.remote_module.write_file(
                '/etc/ceph/%s.client.admin.keyring' % args.cluster,
                keyring,
                0o600,
            )

            distro.conn.exit()

        except RuntimeError as e:
            LOG.error(e)
            errors += 1

    if errors:
        raise exc.GenericError('Failed to configure %d admin hosts' % errors)


@priority(70)
def make(parser):
    """
    Push configuration and client.admin key to a remote host.
    """
    parser.add_argument(
        'client',
        metavar='HOST',
        nargs='+',
        help='host to configure for Ceph administration',
        )
    parser.set_defaults(
        func=admin,
        )



try:
    import configparser
except ImportError:
    import ConfigParser as configparser
import errno
import socket
import os
import shutil
import tempfile
import platform


def platform_information(_linux_distribution=None):
    """ detect platform information from remote host """
    linux_distribution = _linux_distribution or platform.linux_distribution
    distro, release, codename = linux_distribution()
    if not codename and 'debian' in distro.lower():  # this could be an empty string in Debian
        debian_codenames = {
            '8': 'jessie',
            '7': 'wheezy',
            '6': 'squeeze',
        }
        major_version = release.split('.')[0]
        codename = debian_codenames.get(major_version, '')

        # In order to support newer jessie/sid or wheezy/sid strings we test this
        # if sid is buried in the minor, we should use sid anyway.
        if not codename and '/' in release:
            major, minor = release.split('/')
            if minor == 'sid':
                codename = minor
            else:
                codename = major
    if not codename and 'oracle' in distro.lower(): # this could be an empty string in Oracle linux
        codename = 'oracle'
    if not codename and 'virtuozzo linux' in distro.lower(): # this could be an empty string in Virtuozzo linux
        codename = 'virtuozzo'

    return (
        str(distro).rstrip(),
        str(release).rstrip(),
        str(codename).rstrip()
    )


def machine_type():
    """ detect machine type """
    return platform.machine()


def write_sources_list(url, codename, filename='ceph.list', mode=0o644):
    """add deb repo to /etc/apt/sources.list.d/"""
    repo_path = os.path.join('/etc/apt/sources.list.d', filename)
    content = 'deb {url} {codename} main\n'.format(
        url=url,
        codename=codename,
    )
    write_file(repo_path, content.encode('utf-8'), mode)


def write_sources_list_content(content, filename='ceph.list', mode=0o644):
    """add deb repo to /etc/apt/sources.list.d/ from content"""
    repo_path = os.path.join('/etc/apt/sources.list.d', filename)
    if not isinstance(content, str):
        content = content.decode('utf-8')
    write_file(repo_path, content.encode('utf-8'), mode)


def write_yum_repo(content, filename='ceph.repo'):
    """add yum repo file in /etc/yum.repos.d/"""
    repo_path = os.path.join('/etc/yum.repos.d', filename)
    if not isinstance(content, str):
        content = content.decode('utf-8')
    write_file(repo_path, content.encode('utf-8'))


def set_apt_priority(fqdn, path='/etc/apt/preferences.d/ceph.pref'):
    template = "Package: *\nPin: origin {fqdn}\nPin-Priority: 999\n"
    content = template.format(fqdn=fqdn)
    with open(path, 'w') as fout:
        fout.write(content)


def set_repo_priority(sections, path='/etc/yum.repos.d/ceph.repo', priority='1'):
    Config = configparser.ConfigParser()
    Config.read(path)
    Config.sections()
    for section in sections:
        try:
            Config.set(section, 'priority', priority)
        except configparser.NoSectionError:
            # Emperor versions of Ceph used all lowercase sections
            # so lets just try again for the section that failed, maybe
            # we are able to find it if it is lower
            Config.set(section.lower(), 'priority', priority)

    with open(path, 'w') as fout:
        Config.write(fout)

    # And now, because ConfigParser is super duper, we need to remove the
    # assignments so this looks like it was before
    def remove_whitespace_from_assignments():
        separator = "="
        lines = open(path).readlines()
        fp = open(path, "w")
        for line in lines:
            line = line.strip()
            if not line.startswith("#") and separator in line:
                assignment = line.split(separator, 1)
                assignment = tuple(map(str.strip, assignment))
                fp.write("%s%s%s\n" % (assignment[0], separator, assignment[1]))
            else:
                fp.write(line + "\n")

    remove_whitespace_from_assignments()


def write_conf(cluster, conf, overwrite):
    """ write cluster configuration to /etc/ceph/{cluster}.conf """
    path = '/etc/ceph/{cluster}.conf'.format(cluster=cluster)
    tmp_file = tempfile.NamedTemporaryFile('w', dir='/etc/ceph', delete=False)
    err_msg = 'config file %s exists with different content; use --overwrite-conf to overwrite' % path

    if os.path.exists(path):
        with open(path, 'r') as f:
            old = f.read()
            if old != conf and not overwrite:
                raise RuntimeError(err_msg)
        tmp_file.write(conf)
        tmp_file.close()
        shutil.move(tmp_file.name, path)
        os.chmod(path, 0o644)
        return
    if os.path.exists('/etc/ceph'):
        with open(path, 'w') as f:
            f.write(conf)
        os.chmod(path, 0o644)
    else:
        err_msg = '/etc/ceph/ does not exist - could not write config'
        raise RuntimeError(err_msg)


def write_keyring(path, key, uid=-1, gid=-1):
    """ create a keyring file """
    # Note that we *require* to avoid deletion of the temp file
    # otherwise we risk not being able to copy the contents from
    # one file system to the other, hence the `delete=False`
    tmp_file = tempfile.NamedTemporaryFile('wb', delete=False)
    tmp_file.write(key)
    tmp_file.close()
    keyring_dir = os.path.dirname(path)
    if not path_exists(keyring_dir):
        makedir(keyring_dir, uid, gid)
    shutil.move(tmp_file.name, path)


def create_mon_path(path, uid=-1, gid=-1):
    """create the mon path if it does not exist"""
    if not os.path.exists(path):
        os.makedirs(path)
        os.chown(path, uid, gid);


def create_done_path(done_path, uid=-1, gid=-1):
    """create a done file to avoid re-doing the mon deployment"""
    with open(done_path, 'wb'):
        pass
    os.chown(done_path, uid, gid);


def create_init_path(init_path, uid=-1, gid=-1):
    """create the init path if it does not exist"""
    if not os.path.exists(init_path):
        with open(init_path, 'wb'):
            pass
        os.chown(init_path, uid, gid);


def append_to_file(file_path, contents):
    """append contents to file"""
    with open(file_path, 'a') as f:
        f.write(contents)

def path_getuid(path):
    return os.stat(path).st_uid

def path_getgid(path):
    return os.stat(path).st_gid

def readline(path):
    with open(path) as _file:
        return _file.readline().strip('\n')


def path_exists(path):
    return os.path.exists(path)


def get_realpath(path):
    return os.path.realpath(path)


def listdir(path):
    return os.listdir(path)


def makedir(path, ignored=None, uid=-1, gid=-1):
    ignored = ignored or []
    try:
        os.makedirs(path)
    except OSError as error:
        if error.errno in ignored:
            pass
        else:
            # re-raise the original exception
            raise
    else:
        os.chown(path, uid, gid);


def unlink(_file):
    os.unlink(_file)


def write_monitor_keyring(keyring, monitor_keyring, uid=-1, gid=-1):
    """create the monitor keyring file"""
    write_file(keyring, monitor_keyring, 0o600, None, uid, gid)


def write_file(path, content, mode=0o644, directory=None, uid=-1, gid=-1):
    if directory:
        if path.startswith("/"):
            path = path[1:]
        path = os.path.join(directory, path)
    if os.path.exists(path):
        # Delete file in case we are changing its mode
        os.unlink(path)
    with os.fdopen(os.open(path, os.O_WRONLY | os.O_CREAT, mode), 'wb') as f:
        f.write(content)
    os.chown(path, uid, gid)


def touch_file(path):
    with open(path, 'wb') as f:  # noqa
        pass


def get_file(path):
    """ fetch remote file """
    try:
        with open(path, 'rb') as f:
            return f.read()
    except IOError:
        pass


def object_grep(term, file_object):
    for line in file_object.readlines():
        if term in line:
            return True
    return False


def grep(term, file_path):
    # A small grep-like function that will search for a word in a file and
    # return True if it does and False if it does not.

    # Implemented initially to have a similar behavior as the init system
    # detection in Ceph's init scripts::

    #     # detect systemd
    #     # SYSTEMD=0
    #     grep -qs systemd /proc/1/comm && SYSTEMD=1

    # .. note:: Because we intent to be operating in silent mode, we explicitly
    # return ``False`` if the file does not exist.
    if not os.path.isfile(file_path):
        return False

    with open(file_path) as _file:
        return object_grep(term, _file)


def shortname():
    """get remote short hostname"""
    return socket.gethostname().split('.', 1)[0]


def which_service():
    """ locating the `service` executable... """
    # XXX This should get deprecated at some point. For now
    # it just bypasses and uses the new helper.
    return which('service')


def which(executable):
    """find the location of an executable"""
    locations = (
        '/usr/local/bin',
        '/bin',
        '/usr/bin',
        '/usr/local/sbin',
        '/usr/sbin',
        '/sbin',
    )

    for location in locations:
        executable_path = os.path.join(location, executable)
        if os.path.exists(executable_path):
            return executable_path


def make_mon_removed_dir(path, file_name):
    """ move old monitor data """
    try:
        os.makedirs('/var/lib/ceph/mon-removed')
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise
    shutil.move(path, os.path.join('/var/lib/ceph/mon-removed/', file_name))


def safe_mkdir(path, uid=-1, gid=-1):
    """ create path if it doesn't exist """
    try:
        os.mkdir(path)
    except OSError as e:
        if e.errno == errno.EEXIST:
            pass
        else:
            raise
    else:
        os.chown(path, uid, gid)

def safe_makedirs(path, uid=-1, gid=-1):
    """ create path recursively if it doesn't exist """
    try:
        os.makedirs(path)
    except OSError as e:
        if e.errno == errno.EEXIST:
            pass
        else:
            raise
    else:
        os.chown(path, uid, gid)


def zeroing(dev):
    """ zeroing last few blocks of device """
    # this kills the crab
    #
    # sgdisk will wipe out the main copy of the GPT partition
    # table (sorry), but it doesn't remove the backup copies, and
    # subsequent commands will continue to complain and fail when
    # they see those.  zeroing the last few blocks of the device
    # appears to do the trick.
    lba_size = 4096
    size = 33 * lba_size
    return True
    with open(dev, 'wb') as f:
        f.seek(-size, os.SEEK_END)
        f.write(size*b'\0')


def enable_yum_priority_obsoletes(path="/etc/yum/pluginconf.d/priorities.conf"):
    """Configure Yum priorities to include obsoletes"""
    config = configparser.ConfigParser()
    config.read(path)
    config.set('main', 'check_obsoletes', '1')
    with open(path, 'w') as fout:
        config.write(fout)


# remoto magic, needed to execute these functions remotely
if __name__ == '__channelexec__':
    for item in channel:  # noqa
        channel.send(eval(item))  # noqa



import logging
from ceph_deploy.util import templates
from ceph_deploy.lib import remoto
from ceph_deploy.hosts.common import map_components
from ceph_deploy.util.paths import gpg
from ceph_deploy.util import net


LOG = logging.getLogger(__name__)
NON_SPLIT_PACKAGES = ['ceph-osd', 'ceph-mon', 'ceph-mds']


def rpm_dist(distro):
    if distro.normalized_name in ['redhat', 'centos', 'scientific', 'oracle', 'virtuozzo'] and distro.normalized_release.int_major >= 6:
        return 'el' + distro.normalized_release.major
    return 'el6'


def repository_url_part(distro):
    """
    Historically everything CentOS, RHEL, and Scientific has been mapped to
    `el6` urls, but as we are adding repositories for `rhel`, the URLs should
    map correctly to, say, `rhel6` or `rhel7`.

    This function looks into the `distro` object and determines the right url
    part for the given distro, falling back to `el6` when all else fails.

    Specifically to work around the issue of CentOS vs RHEL::

        >>> import platform
        >>> platform.linux_distribution()
        ('Red Hat Enterprise Linux Server', '7.0', 'Maipo')

    """
    if distro.normalized_release.int_major >= 6:
        if distro.normalized_name == 'redhat':
            return 'rhel' + distro.normalized_release.major
        if distro.normalized_name in ['centos', 'scientific', 'oracle', 'virtuozzo']:
            return 'el' + distro.normalized_release.major

    return 'el6'


def install(distro, version_kind, version, adjust_repos, **kw):
    packages = map_components(
        NON_SPLIT_PACKAGES,
        kw.pop('components', [])
    )

    gpgcheck = kw.pop('gpgcheck', 1)
    logger = distro.conn.logger
    machine = distro.machine_type
    repo_part = repository_url_part(distro)
    dist = rpm_dist(distro)

    distro.packager.clean()

    # Get EPEL installed before we continue:
    if adjust_repos:
        distro.packager.install('epel-release')
        distro.packager.install('yum-plugin-priorities')
        distro.conn.remote_module.enable_yum_priority_obsoletes()
        logger.warning('check_obsoletes has been enabled for Yum priorities plugin')
    if version_kind in ['stable', 'testing']:
        key = 'release'
    else:
        key = 'autobuild'

    if adjust_repos:
        if version_kind in ['stable', 'testing']:
            distro.packager.add_repo_gpg_key(gpg.url(key))

            if version_kind == 'stable':
                url = 'https://download.ceph.com/rpm-{version}/{repo}/'.format(
                    version=version,
                    repo=repo_part,
                    )
            elif version_kind == 'testing':
                url = 'https://download.ceph.com/rpm-testing/{repo}/'.format(repo=repo_part)

            remoto.process.run(
                distro.conn,
                [
                    'rpm',
                    '-Uvh',
                    '--replacepkgs',
                    '{url}noarch/ceph-release-1-0.{dist}.noarch.rpm'.format(url=url, dist=dist),
                ],
            )

        elif version_kind in ['dev', 'dev_commit']:
            logger.info('skipping install of ceph-release package')
            logger.info('repo file will be created manually')
            shaman_url = 'https://shaman.ceph.com/api/repos/ceph/{version}/{sha1}/{distro}/{distro_version}/repo/?arch={arch}'.format(
                distro=distro.normalized_name,
                distro_version=distro.normalized_release.major,
                version=kw['args'].dev,
                sha1=kw['args'].dev_commit or 'latest',
                arch=machine
                )
            LOG.debug('fetching repo information from: %s' % shaman_url)
            content = net.get_chacra_repo(shaman_url)
            mirror_install(
                distro,
                '',  # empty repo_url
                None,  # no need to use gpg here, repos are unsigned
                adjust_repos=True,
                extra_installs=False,
                gpgcheck=gpgcheck,
                repo_content=content
            )

        else:
            raise Exception('unrecognized version_kind %s' % version_kind)

        # set the right priority
        logger.warning('ensuring that /etc/yum.repos.d/ceph.repo contains a high priority')
        distro.conn.remote_module.set_repo_priority(['Ceph', 'Ceph-noarch', 'ceph-source'])
        logger.warning('altered ceph.repo priorities to contain: priority=1')

    if packages:
        distro.packager.install(packages)


def mirror_install(distro, repo_url, gpg_url, adjust_repos, extra_installs=True, **kw):
    packages = map_components(
        NON_SPLIT_PACKAGES,
        kw.pop('components', [])
    )
    repo_url = repo_url.strip('/')  # Remove trailing slashes
    gpgcheck = kw.pop('gpgcheck', 1)

    distro.packager.clean()

    if adjust_repos:
        if gpg_url:
            distro.packager.add_repo_gpg_key(gpg_url)

        ceph_repo_content = templates.ceph_repo.format(
            repo_url=repo_url,
            gpg_url=gpg_url,
            gpgcheck=gpgcheck,
        )
        content = kw.get('repo_content', ceph_repo_content)

        distro.conn.remote_module.write_yum_repo(content)
        # set the right priority
        if distro.packager.name == 'yum':
            distro.packager.install('yum-plugin-priorities')
        distro.conn.remote_module.set_repo_priority(['Ceph', 'Ceph-noarch', 'ceph-source'])
        distro.conn.logger.warning('altered ceph.repo priorities to contain: priority=1')


    if extra_installs and packages:
        distro.packager.install(packages)


def repo_install(distro, reponame, baseurl, gpgkey, **kw):
    packages = map_components(
        NON_SPLIT_PACKAGES,
        kw.pop('components', [])
    )
    logger = distro.conn.logger
    # Get some defaults
    name = kw.pop('name', '%s repo' % reponame)
    enabled = kw.pop('enabled', 1)
    gpgcheck = kw.pop('gpgcheck', 1)
    install_ceph = kw.pop('install_ceph', False)
    proxy = kw.pop('proxy', '') # will get ignored if empty
    _type = 'repo-md'
    baseurl = baseurl.strip('/')  # Remove trailing slashes

    distro.packager.clean()

    if gpgkey:
        distro.packager.add_repo_gpg_key(gpgkey)

    repo_content = templates.custom_repo(
        reponame=reponame,
        name=name,
        baseurl=baseurl,
        enabled=enabled,
        gpgcheck=gpgcheck,
        _type=_type,
        gpgkey=gpgkey,
        proxy=proxy,
        **kw
    )

    distro.conn.remote_module.write_yum_repo(
        repo_content,
        "%s.repo" % reponame
    )

    repo_path = '/etc/yum.repos.d/{reponame}.repo'.format(reponame=reponame)

    # set the right priority
    if kw.get('priority'):
        if distro.packager.name == 'yum':
            distro.packager.install('yum-plugin-priorities')

        distro.conn.remote_module.set_repo_priority([reponame], repo_path)
        logger.warning('altered {reponame}.repo priorities to contain: priority=1'.format(
            reponame=reponame)
        )

    # Some custom repos do not need to install ceph
    if install_ceph and packages:
        distro.packager.install(packages)



import ceph_deploy.rgw as rgw


def test_rgw_prefix_auto():
    daemon = rgw.colon_separated("hostname")
    assert daemon == ("hostname", "rgw.hostname")


def test_rgw_prefix_custom():
    daemon = rgw.colon_separated("hostname:mydaemon")
    assert daemon == ("hostname", "rgw.mydaemon")



import argparse
import logging
import os

from ceph_deploy import hosts
from ceph_deploy.cliutil import priority
from ceph_deploy.lib import remoto
from ceph_deploy.util.constants import default_components
from ceph_deploy.util.paths import gpg

LOG = logging.getLogger(__name__)


def sanitize_args(args):
    """
    args may need a bunch of logic to set proper defaults that argparse is
    not well suited for.
    """
    if args.release is None:
        args.release = 'jewel'
        args.default_release = True

    # XXX This whole dance is because --stable is getting deprecated
    if args.stable is not None:
        LOG.warning('the --stable flag is deprecated, use --release instead')
        args.release = args.stable
    # XXX Tango ends here.

    return args


def detect_components(args, distro):
    """
    Since the package split, now there are various different Ceph components to
    install like:

    * ceph
    * ceph-mon
    * ceph-mgr
    * ceph-osd
    * ceph-mds

    This helper function should parse the args that may contain specifics about
    these flags and return the default if none are passed in (which is, install
    everything)
    """
    # the flag that prevents all logic here is the `--repo` flag which is used
    # when no packages should be installed, just the repo files, so check for
    # that here and return an empty list (which is equivalent to say 'no
    # packages should be installed')
    if args.repo:
        return []

    flags = {
        'install_osd': 'ceph-osd',
        'install_rgw': 'ceph-radosgw',
        'install_mds': 'ceph-mds',
        'install_mon': 'ceph-mon',
        'install_mgr': 'ceph-mgr',
        'install_common': 'ceph-common',
        'install_tests': 'ceph-test',
    }

    if distro.is_rpm:
        defaults = default_components.rpm
    else:
        defaults = default_components.deb
        # different naming convention for deb than rpm for radosgw
        flags['install_rgw'] = 'radosgw'

    if args.install_all:
        return defaults
    else:
        components = []
        for k, v in flags.items():
            if getattr(args, k, False):
                components.append(v)
        # if we have some components selected from flags then return that,
        # otherwise return defaults because no flags and no `--repo` means we
        # should get all of them by default
        return components or defaults


def install(args):
    args = sanitize_args(args)

    if args.repo:
        return install_repo(args)

    gpgcheck = 0 if args.nogpgcheck else 1

    if args.version_kind == 'stable':
        version = args.release
    else:
        version = getattr(args, args.version_kind)

    version_str = args.version_kind

    if version:
        version_str += ' version {version}'.format(version=version)
    LOG.debug(
        'Installing %s on cluster %s hosts %s',
        version_str,
        args.cluster,
        ' '.join(args.host),
    )

    for hostname in args.host:
        LOG.debug('Detecting platform for host %s ...', hostname)
        distro = hosts.get(
            hostname,
            username=args.username,
            # XXX this should get removed once ceph packages are split for
            # upstream. If default_release is True, it means that the user is
            # trying to install on a RHEL machine and should expect to get RHEL
            # packages. Otherwise, it will need to specify either a specific
            # version, or repo, or a development branch. Other distro users
            # should not see any differences.
            use_rhceph=args.default_release,
            )
        LOG.info(
            'Distro info: %s %s %s',
            distro.name,
            distro.release,
            distro.codename
        )

        components = detect_components(args, distro)
        if distro.init == 'sysvinit' and args.cluster != 'ceph':
            LOG.error('refusing to install on host: %s, with custom cluster name: %s' % (
                    hostname,
                    args.cluster,
                )
            )
            LOG.error('custom cluster names are not supported on sysvinit hosts')
            continue

        rlogger = logging.getLogger(hostname)
        rlogger.info('installing Ceph on %s' % hostname)

        cd_conf = getattr(args, 'cd_conf', None)

        # custom repo arguments
        repo_url = os.environ.get('CEPH_DEPLOY_REPO_URL') or args.repo_url
        gpg_url = os.environ.get('CEPH_DEPLOY_GPG_URL') or args.gpg_url
        gpg_fallback = gpg.url('release')

        if gpg_url is None and repo_url:
            LOG.warning('--gpg-url was not used, will fallback')
            LOG.warning('using GPG fallback: %s', gpg_fallback)
            gpg_url = gpg_fallback

        if args.local_mirror:
            if args.username:
                hostname = "%s@%s" % (args.username, hostname)
            remoto.rsync(hostname, args.local_mirror, '/opt/ceph-deploy/repo', distro.conn.logger, sudo=True)
            repo_url = 'file:///opt/ceph-deploy/repo'
            gpg_url = 'file:///opt/ceph-deploy/repo/release.asc'

        if repo_url:  # triggers using a custom repository
            # the user used a custom repo url, this should override anything
            # we can detect from the configuration, so warn about it
            if cd_conf:
                if cd_conf.get_default_repo():
                    rlogger.warning('a default repo was found but it was \
                        overridden on the CLI')
                if args.release in cd_conf.get_repos():
                    rlogger.warning('a custom repo was found but it was \
                        overridden on the CLI')

            rlogger.info('using custom repository location: %s', repo_url)
            distro.mirror_install(
                distro,
                repo_url,
                gpg_url,
                args.adjust_repos,
                components=components,
                gpgcheck=gpgcheck,
                args=args
            )

        # Detect and install custom repos here if needed
        elif should_use_custom_repo(args, cd_conf, repo_url):
            LOG.info('detected valid custom repositories from config file')
            custom_repo(distro, args, cd_conf, rlogger)

        else:  # otherwise a normal installation
            distro.install(
                distro,
                args.version_kind,
                version,
                args.adjust_repos,
                components=components,
                gpgcheck = gpgcheck,
                args=args
            )

        # Check the ceph version we just installed
        hosts.common.ceph_version(distro.conn)
        distro.conn.exit()


def should_use_custom_repo(args, cd_conf, repo_url):
    """
    A boolean to determine the logic needed to proceed with a custom repo
    installation instead of cramming everything nect to the logic operator.
    """
    if repo_url:
        # repo_url signals a CLI override, return False immediately
        return False
    if cd_conf:
        if cd_conf.has_repos:
            has_valid_release = args.release in cd_conf.get_repos()
            has_default_repo = cd_conf.get_default_repo()
            if has_valid_release or has_default_repo:
                return True
    return False


def custom_repo(distro, args, cd_conf, rlogger, install_ceph=None):
    """
    A custom repo install helper that will go through config checks to retrieve
    repos (and any extra repos defined) and install those

    ``cd_conf`` is the object built from argparse that holds the flags and
    information needed to determine what metadata from the configuration to be
    used.
    """
    default_repo = cd_conf.get_default_repo()
    components = detect_components(args, distro)
    if args.release in cd_conf.get_repos():
        LOG.info('will use repository from conf: %s' % args.release)
        default_repo = args.release
    elif default_repo:
        LOG.info('will use default repository: %s' % default_repo)

    # At this point we know there is a cd_conf and that it has custom
    # repos make sure we were able to detect and actual repo
    if not default_repo:
        LOG.warning('a ceph-deploy config was found with repos \
            but could not default to one')
    else:
        options = dict(cd_conf.items(default_repo))
        options['install_ceph'] = False if install_ceph is False else True
        extra_repos = cd_conf.get_list(default_repo, 'extra-repos')
        rlogger.info('adding custom repository file')
        try:
            distro.repo_install(
                distro,
                default_repo,
                options.pop('baseurl'),
                options.pop('gpgkey'),
                components=components,
                **options
            )
        except KeyError as err:
            raise RuntimeError('missing required key: %s in config section: %s' % (err, default_repo))

        for xrepo in extra_repos:
            rlogger.info('adding extra repo file: %s.repo' % xrepo)
            options = dict(cd_conf.items(xrepo))
            try:
                distro.repo_install(
                    distro,
                    xrepo,
                    options.pop('baseurl'),
                    options.pop('gpgkey'),
                    components=components,
                    **options
                )
            except KeyError as err:
                raise RuntimeError('missing required key: %s in config section: %s' % (err, xrepo))


def install_repo(args):
    """
    For a user that only wants to install the repository only (and avoid
    installing Ceph and its dependencies).
    """
    cd_conf = getattr(args, 'cd_conf', None)

    for hostname in args.host:
        LOG.debug('Detecting platform for host %s ...', hostname)
        distro = hosts.get(
            hostname,
            username=args.username,
            # XXX this should get removed once Ceph packages are split for
            # upstream. If default_release is True, it means that the user is
            # trying to install on a RHEL machine and should expect to get RHEL
            # packages. Otherwise, it will need to specify either a specific
            # version, or repo, or a development branch. Other distro users should
            # not see any differences.
            use_rhceph=args.default_release,
        )
        rlogger = logging.getLogger(hostname)

        LOG.info(
            'Distro info: %s %s %s',
            distro.name,
            distro.release,
            distro.codename
        )

        custom_repo(distro, args, cd_conf, rlogger, install_ceph=False)

def remove(args, purge):
    LOG.info('note that some dependencies *will not* be removed because they can cause issues with qemu-kvm')
    LOG.info('like: librbd1 and librados2')
    remove_action = 'Uninstalling'
    if purge:
        remove_action = 'Purging'
    LOG.debug(
        '%s on cluster %s hosts %s',
        remove_action,
        args.cluster,
        ' '.join(args.host),
        )

    for hostname in args.host:
        LOG.debug('Detecting platform for host %s ...', hostname)

        distro = hosts.get(
            hostname,
            username=args.username,
            use_rhceph=True)
        LOG.info('Distro info: %s %s %s', distro.name, distro.release, distro.codename)
        rlogger = logging.getLogger(hostname)
        rlogger.info('%s Ceph on %s' % (remove_action, hostname))
        distro.uninstall(distro, purge=purge)
        distro.conn.exit()

def uninstall(args):
    remove(args, False)

def purge(args):
    remove(args, True)

def purgedata(args):
    LOG.debug(
        'Purging data from cluster %s hosts %s',
        args.cluster,
        ' '.join(args.host),
        )

    installed_hosts = []
    for hostname in args.host:
        distro = hosts.get(hostname, username=args.username)
        ceph_is_installed = distro.conn.remote_module.which('ceph')
        if ceph_is_installed:
            installed_hosts.append(hostname)
        distro.conn.exit()

    if installed_hosts:
        LOG.error("Ceph is still installed on: %s", installed_hosts)
        raise RuntimeError("refusing to purge data while Ceph is still installed")

    for hostname in args.host:
        distro = hosts.get(hostname, username=args.username)
        LOG.info(
            'Distro info: %s %s %s',
            distro.name,
            distro.release,
            distro.codename
        )

        rlogger = logging.getLogger(hostname)
        rlogger.info('purging data on %s' % hostname)

        # Try to remove the contents of /var/lib/ceph first, don't worry
        # about errors here, we deal with them later on
        remoto.process.check(
            distro.conn,
            [
                'rm', '-rf', '--one-file-system', '--', '/var/lib/ceph',
            ]
        )

        # If we failed in the previous call, then we probably have OSDs
        # still mounted, so we unmount them here
        if distro.conn.remote_module.path_exists('/var/lib/ceph'):
            rlogger.warning(
                'OSDs may still be mounted, trying to unmount them'
            )
            remoto.process.run(
                distro.conn,
                [
                    'find', '/var/lib/ceph',
                    '-mindepth', '1',
                    '-maxdepth', '2',
                    '-type', 'd',
                    '-exec', 'umount', '{}', ';',
                ]
            )

            # And now we try again to remove the contents, since OSDs should be
            # unmounted, but this time we do check for errors
            remoto.process.run(
                distro.conn,
                [
                    'rm', '-rf', '--one-file-system', '--', '/var/lib/ceph',
                ]
            )

        remoto.process.run(
            distro.conn,
            [
                'rm', '-rf', '--one-file-system', '--', '/etc/ceph/',
            ]
        )

        distro.conn.exit()


class StoreVersion(argparse.Action):
    """
    Like ``"store"`` but also remember which one of the exclusive
    options was set.

    There are three kinds of versions: stable, testing and dev.
    This sets ``version_kind`` to be the right one of the above.

    This kludge essentially lets us differentiate explicitly set
    values from defaults.
    """
    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, values)
        if self.dest == 'release':
            self.dest = 'stable'
        namespace.version_kind = self.dest


@priority(20)
def make(parser):
    """
    Install Ceph packages on remote hosts.
    """

    version = parser.add_mutually_exclusive_group()

    # XXX deprecated in favor of release
    version.add_argument(
        '--stable',
        nargs='?',
        action=StoreVersion,
        metavar='CODENAME',
        help='[DEPRECATED] install a release known as CODENAME\
                (done by default) (default: %(default)s)',
    )

    version.add_argument(
        '--release',
        nargs='?',
        action=StoreVersion,
        metavar='CODENAME',
        help='install a release known as CODENAME\
                (done by default) (default: %(default)s)',
    )

    version.add_argument(
        '--testing',
        nargs=0,
        action=StoreVersion,
        help='install the latest development release',
    )

    version.add_argument(
        '--dev',
        nargs='?',
        action=StoreVersion,
        const='master',
        metavar='BRANCH_OR_TAG',
        help='install a bleeding edge build from Git branch\
                or tag (default: %(default)s)',
    )
    parser.add_argument(
        '--dev-commit',
        nargs='?',
        action=StoreVersion,
        metavar='COMMIT',
        help='install a bleeding edge build from Git commit (defaults to master branch)',
    )

    version.set_defaults(
        stable=None,  # XXX deprecated in favor of release
        release=None,  # Set the default release in sanitize_args()
        dev='master',
        version_kind='stable',
    )

    parser.add_argument(
        '--mon',
        dest='install_mon',
        action='store_true',
        help='install the mon component only',
    )

    parser.add_argument(
        '--mgr',
        dest='install_mgr',
        action='store_true',
        help='install the mgr component only',
    )

    parser.add_argument(
        '--mds',
        dest='install_mds',
        action='store_true',
        help='install the mds component only',
    )

    parser.add_argument(
        '--rgw',
        dest='install_rgw',
        action='store_true',
        help='install the rgw component only',
    )

    parser.add_argument(
        '--osd',
        dest='install_osd',
        action='store_true',
        help='install the osd component only',
    )

    parser.add_argument(
        '--tests',
        dest='install_tests',
        action='store_true',
        help='install the testing components',
    )

    parser.add_argument(
        '--cli', '--common',
        dest='install_common',
        action='store_true',
        help='install the common component only',
    )

    parser.add_argument(
        '--all',
        dest='install_all',
        action='store_true',
        help='install all Ceph components (mon, osd, mds, rgw) except tests. This is the default',
    )

    repo = parser.add_mutually_exclusive_group()

    repo.add_argument(
        '--adjust-repos',
        dest='adjust_repos',
        action='store_true',
        help='install packages modifying source repos',
    )

    repo.add_argument(
        '--no-adjust-repos',
        dest='adjust_repos',
        action='store_false',
        help='install packages without modifying source repos',
    )

    repo.add_argument(
        '--repo',
        action='store_true',
        help='install repo files only (skips package installation)',
    )

    repo.set_defaults(
        adjust_repos=True,
    )

    parser.add_argument(
        'host',
        metavar='HOST',
        nargs='+',
        help='hosts to install on',
    )

    parser.add_argument(
        '--local-mirror',
        nargs='?',
        const='PATH',
        default=None,
        help='Fetch packages and push them to hosts for a local repo mirror',
    )

    parser.add_argument(
        '--repo-url',
        nargs='?',
        dest='repo_url',
        help='specify a repo URL that mirrors/contains Ceph packages',
    )

    parser.add_argument(
        '--gpg-url',
        nargs='?',
        dest='gpg_url',
        help='specify a GPG key URL to be used with custom repos\
                (defaults to ceph.com)'
    )

    parser.add_argument(
        '--nogpgcheck',
        action='store_true',
        help='install packages without gpgcheck',
    )

    parser.set_defaults(
        func=install,
    )


@priority(80)
def make_uninstall(parser):
    """
    Remove Ceph packages from remote hosts.
    """
    parser.add_argument(
        'host',
        metavar='HOST',
        nargs='+',
        help='hosts to uninstall Ceph from',
        )
    parser.set_defaults(
        func=uninstall,
        )


@priority(80)
def make_purge(parser):
    """
    Remove Ceph packages from remote hosts and purge all data.
    """
    parser.add_argument(
        'host',
        metavar='HOST',
        nargs='+',
        help='hosts to purge Ceph from',
        )
    parser.set_defaults(
        func=purge,
        )


@priority(80)
def make_purge_data(parser):
    """
    Purge (delete, destroy, discard, shred) any Ceph data from /var/lib/ceph
    """
    parser.add_argument(
        'host',
        metavar='HOST',
        nargs='+',
        help='hosts to purge Ceph data from',
        )
    parser.set_defaults(
        func=purgedata,
        )



import pkg_resources
import argparse
import logging
import textwrap
import os
import sys

import ceph_deploy
from ceph_deploy import exc, validate
from ceph_deploy.util import log
from ceph_deploy.util.decorators import catches

LOG = logging.getLogger(__name__)


__header__ = textwrap.dedent("""
    -^-
   /   \\
   |O o|  ceph-deploy v%s
   ).-.(
  '/|||\`
  | '|` |
    '|`

Full documentation can be found at: http://ceph.com/ceph-deploy/docs
""" % ceph_deploy.__version__)


def log_flags(args, logger=None):
    logger = logger or LOG
    logger.info('ceph-deploy options:')

    for k, v in args.__dict__.items():
        if k.startswith('_'):
            continue
        logger.info(' %-30s: %s' % (k, v))


def get_parser():
    parser = argparse.ArgumentParser(
        prog='ceph-deploy',
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description='Easy Ceph deployment\n\n%s' % __header__,
        )
    verbosity = parser.add_mutually_exclusive_group(required=False)
    verbosity.add_argument(
        '-v', '--verbose',
        action='store_true', dest='verbose', default=False,
        help='be more verbose',
        )
    verbosity.add_argument(
        '-q', '--quiet',
        action='store_true', dest='quiet',
        help='be less verbose',
        )
    parser.add_argument(
        '--version',
        action='version',
        version='%s' % ceph_deploy.__version__,
        help='the current installed version of ceph-deploy',
        )
    parser.add_argument(
        '--username',
        help='the username to connect to the remote host',
        )
    parser.add_argument(
        '--overwrite-conf',
        action='store_true',
        help='overwrite an existing conf file on remote host (if present)',
        )
    parser.add_argument(
        '--cluster',
        metavar='NAME',
        help='name of the cluster',
        type=validate.alphanumeric,
        )
    parser.add_argument(
        '--ceph-conf',
        dest='ceph_conf',
        help='use (or reuse) a given ceph.conf file',
    )
    sub = parser.add_subparsers(
        title='commands',
        metavar='COMMAND',
        help='description',
        )
    sub.required = True
    entry_points = [
        (ep.name, ep.load())
        for ep in pkg_resources.iter_entry_points('ceph_deploy.cli')
        ]
    entry_points.sort(
        key=lambda name_fn: getattr(name_fn[1], 'priority', 100),
        )
    for (name, fn) in entry_points:
        p = sub.add_parser(
            name,
            description=fn.__doc__,
            help=fn.__doc__,
            )
        if not os.environ.get('CEPH_DEPLOY_TEST'):
            p.set_defaults(cd_conf=ceph_deploy.conf.cephdeploy.load())

        # flag if the default release is being used
        p.set_defaults(default_release=False)
        fn(p)
        p.required = True
    parser.set_defaults(
        cluster='ceph',
        )

    return parser


@catches((KeyboardInterrupt, RuntimeError, exc.DeployError,), handle_all=True)
def _main(args=None, namespace=None):
    # Set console logging first with some defaults, to prevent having exceptions
    # before hitting logging configuration. The defaults can/will get overridden
    # later.

    # Console Logger
    sh = logging.StreamHandler()
    sh.setFormatter(log.color_format())
    sh.setLevel(logging.WARNING)

    # because we're in a module already, __name__ is not the ancestor of
    # the rest of the package; use the root as the logger for everyone
    root_logger = logging.getLogger()

    # allow all levels at root_logger, handlers control individual levels
    root_logger.setLevel(logging.DEBUG)
    root_logger.addHandler(sh)

    parser = get_parser()
    if len(sys.argv) < 2:
        parser.print_help()
        sys.exit()
    else:
        args = parser.parse_args(args=args, namespace=namespace)

    console_loglevel = logging.DEBUG  # start at DEBUG for now
    if args.quiet:
        console_loglevel = logging.WARNING
    if args.verbose:
        console_loglevel = logging.DEBUG

    # Console Logger
    sh.setLevel(console_loglevel)

    # File Logger
    fh = logging.FileHandler('ceph-deploy-{cluster}.log'.format(cluster=args.cluster))
    fh.setLevel(logging.DEBUG)
    fh.setFormatter(logging.Formatter(log.FILE_FORMAT))

    root_logger.addHandler(fh)

    # Reads from the config file and sets values for the global
    # flags and the given sub-command
    # the one flag that will never work regardless of the config settings is
    # logging because we cannot set it before hand since the logging config is
    # not ready yet. This is the earliest we can do.
    args = ceph_deploy.conf.cephdeploy.set_overrides(args)

    LOG.info("Invoked (%s): %s" % (
        ceph_deploy.__version__,
        ' '.join(sys.argv))
    )
    log_flags(args)

    return args.func(args)


def main(args=None, namespace=None):
    try:
        _main(args=args, namespace=namespace)
    finally:
        # This block is crucial to avoid having issues with
        # Python spitting non-sense thread exceptions. We have already
        # handled what we could, so close stderr and stdout.
        if not os.environ.get('CEPH_DEPLOY_TEST'):
            try:
                sys.stdout.close()
            except:
                pass
            try:
                sys.stderr.close()
            except:
                pass



try:
    from urllib.request import urlopen
    from urllib.error import HTTPError
except ImportError:
    from urllib2 import urlopen, HTTPError

from ceph_deploy import exc
import logging
import re
import socket
from ceph_deploy.lib import remoto


LOG = logging.getLogger(__name__)


# TODO: at some point, it might be way more accurate to do this in the actual
# host where we need to get IPs from. SaltStack does this by calling `ip` and
# parsing the output, which is probably the one true way of dealing with it.

def get_nonlocal_ip(host, subnet=None):
    """
    Search result of getaddrinfo() for a non-localhost-net address
    """
    try:
        ailist = socket.getaddrinfo(host, None)
    except socket.gaierror:
        raise exc.UnableToResolveError(host)
    for ai in ailist:
        # an ai is a 5-tuple; the last element is (ip, port)
        ip = ai[4][0]
        if subnet and ip_in_subnet(ip, subnet):
            LOG.info('found ip (%s) for host (%s) to be in cluster subnet (%s)' % (
                ip,
                host,
                subnet,)
            )

            return ip

        if not ip.startswith('127.'):
            if subnet:
                LOG.warning('could not match ip (%s) for host (%s) for cluster subnet (%s)' % (
                    ip,
                    host,
                    subnet,)
                )
            return ip
    raise exc.UnableToResolveError(host)


def ip_in_subnet(ip, subnet):
    """Does IP exists in a given subnet utility. Returns a boolean"""
    ipaddr = int(''.join(['%02x' % int(x) for x in ip.split('.')]), 16)
    netstr, bits = subnet.split('/')
    netaddr = int(''.join(['%02x' % int(x) for x in netstr.split('.')]), 16)
    mask = (0xffffffff << (32 - int(bits))) & 0xffffffff
    return (ipaddr & mask) == (netaddr & mask)


def in_subnet(cidr, addrs=None):
    """
    Returns True if host is within specified subnet, otherwise False
    """
    for address in addrs:
        if ip_in_subnet(address, cidr):
            return True
    return False


def ip_addresses(conn, interface=None, include_loopback=False):
    """
    Returns a list of IPv4 addresses assigned to the host. 127.0.0.1 is
    ignored, unless 'include_loopback=True' is indicated. If 'interface' is
    provided, then only IP addresses from that interface will be returned.

    Example output looks like::

        >>> ip_addresses(conn)
        >>> ['192.168.1.111', '10.0.1.12']

    """
    ret = set()
    ifaces = linux_interfaces(conn)
    if interface is None:
        target_ifaces = ifaces
    else:
        target_ifaces = dict((k, v) for k, v in ifaces.items()
                             if k == interface)
        if not target_ifaces:
            LOG.error('Interface {0} not found.'.format(interface))
    for ipv4_info in target_ifaces.values():
        for ipv4 in ipv4_info.get('inet', []):
            loopback = in_subnet('127.0.0.0/8', [ipv4.get('address')]) or ipv4.get('label') == 'lo'
            if not loopback or include_loopback:
                ret.add(ipv4['address'])
        for secondary in ipv4_info.get('secondary', []):
            addr = secondary.get('address')
            if addr and secondary.get('type') == 'inet':
                if include_loopback or (not include_loopback and not in_subnet('127.0.0.0/8', [addr])):
                    ret.add(addr)
    if ret:
        conn.logger.debug('IP addresses found: %s' % str(list(ret)))
    return sorted(list(ret))


def linux_interfaces(conn):
    """
    Obtain interface information for *NIX/BSD variants in remote servers.

    Example output from a remote node with a couple of interfaces::

        {'eth0': {'hwaddr': '08:00:27:08:c2:e4',
                  'inet': [{'address': '10.0.2.15',
                            'broadcast': '10.0.2.255',
                            'label': 'eth0',
                            'netmask': '255.255.255.0'}],
                  'inet6': [{'address': 'fe80::a00:27ff:fe08:c2e4',
                             'prefixlen': '64'}],
                  'up': True},
         'eth1': {'hwaddr': '08:00:27:70:06:f1',
                  'inet': [{'address': '192.168.111.101',
                            'broadcast': '192.168.111.255',
                            'label': 'eth1',
                            'netmask': '255.255.255.0'}],
                  'inet6': [{'address': 'fe80::a00:27ff:fe70:6f1',
                             'prefixlen': '64'}],
                  'up': True},
         'lo': {'hwaddr': '00:00:00:00:00:00',
                'inet': [{'address': '127.0.0.1',
                          'broadcast': None,
                          'label': 'lo',
                          'netmask': '255.0.0.0'}],
                'inet6': [{'address': '::1', 'prefixlen': '128'}],
                'up': True}}

    :param conn: A connection object to a remote node
    """
    ifaces = dict()
    ip_path = conn.remote_module.which('ip')
    ifconfig_path = None if ip_path else conn.remote_module.which('ifconfig')
    if ip_path:
        cmd1, _, _ = remoto.process.check(
            conn,
            [
                '{0}'.format(ip_path),
                'link',
                'show',
            ],
        )
        cmd2, _, _ = remoto.process.check(
            conn,
            [
                '{0}'.format(ip_path),
                'addr',
                'show',
            ],
        )
        ifaces = _interfaces_ip(b'\n'.join(cmd1).decode('utf-8') + '\n' +
                                b'\n'.join(cmd2).decode('utf-8'))
    elif ifconfig_path:
        cmd, _, _ = remoto.process.check(
            conn,
            [
                '{0}'.format(ifconfig_path),
                '-a',
            ]
        )
        ifaces = _interfaces_ifconfig('\n'.join(cmd))
    return ifaces


def _interfaces_ip(out):
    """
    Uses ip to return a dictionary of interfaces with various information about
    each (up/down state, ip address, netmask, and hwaddr)
    """
    ret = dict()

    def parse_network(value, cols):
        """
        Return a tuple of ip, netmask, broadcast
        based on the current set of cols
        """
        brd = None
        if '/' in value:  # we have a CIDR in this address
            ip, cidr = value.split('/')  # pylint: disable=C0103
        else:
            ip = value  # pylint: disable=C0103
            cidr = 32

        if type_ == 'inet':
            mask = cidr_to_ipv4_netmask(int(cidr))
            if 'brd' in cols:
                brd = cols[cols.index('brd') + 1]
        elif type_ == 'inet6':
            mask = cidr
        return (ip, mask, brd)

    groups = re.compile('\r?\n\\d').split(out)
    for group in groups:
        iface = None
        data = dict()

        for line in group.splitlines():
            if ' ' not in line:
                continue
            match = re.match(r'^\d*:\s+([\w.\-]+)(?:@)?([\w.\-]+)?:\s+<(.+)>', line)
            if match:
                iface, parent, attrs = match.groups()
                if 'UP' in attrs.split(','):
                    data['up'] = True
                else:
                    data['up'] = False
                if parent:
                    data['parent'] = parent
                continue

            cols = line.split()
            if len(cols) >= 2:
                type_, value = tuple(cols[0:2])
                iflabel = cols[-1:][0]
                if type_ in ('inet', 'inet6'):
                    if 'secondary' not in cols:
                        ipaddr, netmask, broadcast = parse_network(value, cols)
                        if type_ == 'inet':
                            if 'inet' not in data:
                                data['inet'] = list()
                            addr_obj = dict()
                            addr_obj['address'] = ipaddr
                            addr_obj['netmask'] = netmask
                            addr_obj['broadcast'] = broadcast
                            addr_obj['label'] = iflabel
                            data['inet'].append(addr_obj)
                        elif type_ == 'inet6':
                            if 'inet6' not in data:
                                data['inet6'] = list()
                            addr_obj = dict()
                            addr_obj['address'] = ipaddr
                            addr_obj['prefixlen'] = netmask
                            data['inet6'].append(addr_obj)
                    else:
                        if 'secondary' not in data:
                            data['secondary'] = list()
                        ip_, mask, brd = parse_network(value, cols)
                        data['secondary'].append({
                            'type': type_,
                            'address': ip_,
                            'netmask': mask,
                            'broadcast': brd,
                            'label': iflabel,
                        })
                        del ip_, mask, brd
                elif type_.startswith('link'):
                    data['hwaddr'] = value
        if iface:
            ret[iface] = data
            del iface, data
    return ret


def _interfaces_ifconfig(out):
    """
    Uses ifconfig to return a dictionary of interfaces with various information
    about each (up/down state, ip address, netmask, and hwaddr)
    """
    ret = dict()

    piface = re.compile(r'^([^\s:]+)')
    pmac = re.compile('.*?(?:HWaddr|ether|address:|lladdr) ([0-9a-fA-F:]+)')
    pip = re.compile(r'.*?(?:inet addr:|inet )(.*?)\s')
    pip6 = re.compile('.*?(?:inet6 addr: (.*?)/|inet6 )([0-9a-fA-F:]+)')
    pmask = re.compile(r'.*?(?:Mask:|netmask )(?:((?:0x)?[0-9a-fA-F]{8})|([\d\.]+))')
    pmask6 = re.compile(r'.*?(?:inet6 addr: [0-9a-fA-F:]+/(\d+)|prefixlen (\d+)).*')
    pupdown = re.compile('UP')
    pbcast = re.compile(r'.*?(?:Bcast:|broadcast )([\d\.]+)')

    groups = re.compile('\r?\n(?=\\S)').split(out)
    for group in groups:
        data = dict()
        iface = ''
        updown = False
        for line in group.splitlines():
            miface = piface.match(line)
            mmac = pmac.match(line)
            mip = pip.match(line)
            mip6 = pip6.match(line)
            mupdown = pupdown.search(line)
            if miface:
                iface = miface.group(1)
            if mmac:
                data['hwaddr'] = mmac.group(1)
            if mip:
                if 'inet' not in data:
                    data['inet'] = list()
                addr_obj = dict()
                addr_obj['address'] = mip.group(1)
                mmask = pmask.match(line)
                if mmask:
                    if mmask.group(1):
                        mmask = _number_of_set_bits_to_ipv4_netmask(
                            int(mmask.group(1), 16))
                    else:
                        mmask = mmask.group(2)
                    addr_obj['netmask'] = mmask
                mbcast = pbcast.match(line)
                if mbcast:
                    addr_obj['broadcast'] = mbcast.group(1)
                data['inet'].append(addr_obj)
            if mupdown:
                updown = True
            if mip6:
                if 'inet6' not in data:
                    data['inet6'] = list()
                addr_obj = dict()
                addr_obj['address'] = mip6.group(1) or mip6.group(2)
                mmask6 = pmask6.match(line)
                if mmask6:
                    addr_obj['prefixlen'] = mmask6.group(1) or mmask6.group(2)
                data['inet6'].append(addr_obj)
        data['up'] = updown
        ret[iface] = data
        del data
    return ret


def _number_of_set_bits_to_ipv4_netmask(set_bits):  # pylint: disable=C0103
    """
    Returns an IPv4 netmask from the integer representation of that mask.

    Ex. 0xffffff00 -> '255.255.255.0'
    """
    return cidr_to_ipv4_netmask(_number_of_set_bits(set_bits))


def _number_of_set_bits(x):
    """
    Returns the number of bits that are set in a 32bit int
    """
    #  Taken from http://stackoverflow.com/a/4912729. Many thanks!
    x -= (x >> 1) & 0x55555555
    x = ((x >> 2) & 0x33333333) + (x & 0x33333333)
    x = ((x >> 4) + x) & 0x0f0f0f0f
    x += x >> 8
    x += x >> 16
    return x & 0x0000003f


def cidr_to_ipv4_netmask(cidr_bits):
    """
    Returns an IPv4 netmask
    """
    try:
        cidr_bits = int(cidr_bits)
        if not 1 <= cidr_bits <= 32:
            return ''
    except ValueError:
        return ''

    netmask = ''
    for idx in range(4):
        if idx:
            netmask += '.'
        if cidr_bits >= 8:
            netmask += '255'
            cidr_bits -= 8
        else:
            netmask += '{0:d}'.format(256 - (2 ** (8 - cidr_bits)))
            cidr_bits = 0
    return netmask


def get_request(url):
    try:
        return urlopen(url)
    except HTTPError as err:
        LOG.error('repository might not be available yet')
        raise RuntimeError('%s, failed to fetch %s' % (err, url))


def get_chacra_repo(shaman_url):
    """
    From a Shaman URL, get the chacra url for a repository, read the
    contents that point to the repo and return it as a string.
    """
    shaman_response = get_request(shaman_url)
    chacra_url = shaman_response.geturl()
    chacra_response = get_request(chacra_url)

    return chacra_response.read()



from ceph_deploy import gatherkeys
from ceph_deploy import new
import tempfile
import shutil
import pytest


def write_key_mon_with_caps(path, secret):
    mon_keyring = '[mon.]\nkey = %s\ncaps mon = allow *\n' % secret
    with open(path, 'w', 0o600) as f:
        f.write(mon_keyring)


def write_key_mon_with_caps_with_tab(path, secret):
    mon_keyring = '[mon.]\n\tkey = %s\n\tcaps mon = allow *\n' % secret
    with open(path, 'w', 0o600) as f:
        f.write(mon_keyring)


def write_key_mon_with_caps_with_tab_quote(path, secret):
    mon_keyring = '[mon.]\n\tkey = %s\n\tcaps mon = "allow *"\n' % secret
    with open(path, 'w', 0o600) as f:
        f.write(mon_keyring)


def write_key_mon_without_caps(path, secret):
    mon_keyring = '[mon.]\nkey = %s\n' % secret
    with open(path, 'w', 0o600) as f:
        f.write(mon_keyring)


class TestKeysEquivalent(object):
    """
    Since we are testing things that effect the content of the current working
    directory we should test in a clean empty directory.
    """
    def setup(self):
        """
        Make temp directory for tests.
        """
        self.test_dir = tempfile.mkdtemp()


    def teardown(self):
        """
        Remove temp directory and content
        """
        shutil.rmtree(self.test_dir)


    def test_identical_with_caps(self):
        secret_01 = new.generate_auth_key()
        key_path_01 = self.test_dir + "/01.keyring"
        key_path_02 = self.test_dir + "/02.keyring"
        write_key_mon_with_caps(key_path_01, secret_01)
        write_key_mon_with_caps(key_path_02, secret_01)
        same = gatherkeys._keyring_equivalent(key_path_01, key_path_02)
        assert same is True


    def test_different_with_caps(self):
        secret_01 = new.generate_auth_key()
        secret_02 = new.generate_auth_key()
        key_path_01 = self.test_dir + "/01.keyring"
        key_path_02 = self.test_dir + "/02.keyring"
        write_key_mon_with_caps(key_path_01, secret_01)
        write_key_mon_with_caps(key_path_02, secret_02)
        same = gatherkeys._keyring_equivalent(key_path_01, key_path_02)
        assert same is False


    def test_identical_without_caps(self):
        secret_01 = new.generate_auth_key()
        key_path_01 = self.test_dir + "/01.keyring"
        key_path_02 = self.test_dir + "/02.keyring"
        write_key_mon_without_caps(key_path_01, secret_01)
        write_key_mon_without_caps(key_path_02, secret_01)
        same = gatherkeys._keyring_equivalent(key_path_01, key_path_02)
        assert same is True


    def test_different_without_caps(self):
        secret_01 = new.generate_auth_key()
        secret_02 = new.generate_auth_key()
        key_path_01 = self.test_dir + "/01.keyring"
        key_path_02 = self.test_dir + "/02.keyring"
        write_key_mon_without_caps(key_path_01, secret_01)
        write_key_mon_without_caps(key_path_02, secret_02)
        same = gatherkeys._keyring_equivalent(key_path_01, key_path_02)
        assert same is False


    def test_identical_mixed_caps(self):
        secret_01 = new.generate_auth_key()
        key_path_01 = self.test_dir + "/01.keyring"
        key_path_02 = self.test_dir + "/02.keyring"
        write_key_mon_with_caps(key_path_01, secret_01)
        write_key_mon_without_caps(key_path_02, secret_01)
        same = gatherkeys._keyring_equivalent(key_path_01, key_path_02)
        assert same is True


    def test_different_mixed_caps(self):
        secret_01 = new.generate_auth_key()
        secret_02 = new.generate_auth_key()
        key_path_01 = self.test_dir + "/01.keyring"
        key_path_02 = self.test_dir + "/02.keyring"
        write_key_mon_with_caps(key_path_01, secret_01)
        write_key_mon_without_caps(key_path_02, secret_02)
        same = gatherkeys._keyring_equivalent(key_path_01, key_path_02)
        assert same is False


    def test_identical_caps_mixed_tabs(self):
        secret_01 = new.generate_auth_key()
        key_path_01 = self.test_dir + "/01.keyring"
        key_path_02 = self.test_dir + "/02.keyring"
        write_key_mon_with_caps(key_path_01, secret_01)
        write_key_mon_with_caps_with_tab(key_path_02, secret_01)
        same = gatherkeys._keyring_equivalent(key_path_01, key_path_02)
        assert same is True


    def test_different_caps_mixed_tabs(self):
        secret_01 = new.generate_auth_key()
        secret_02 = new.generate_auth_key()
        key_path_01 = self.test_dir + "/01.keyring"
        key_path_02 = self.test_dir + "/02.keyring"
        write_key_mon_with_caps(key_path_01, secret_01)
        write_key_mon_with_caps_with_tab(key_path_02, secret_02)
        same = gatherkeys._keyring_equivalent(key_path_01, key_path_02)
        assert same is False


    def test_identical_caps_mixed_quote(self):
        secret_01 = new.generate_auth_key()
        key_path_01 = self.test_dir + "/01.keyring"
        key_path_02 = self.test_dir + "/02.keyring"
        write_key_mon_with_caps_with_tab(key_path_01, secret_01)
        write_key_mon_with_caps_with_tab_quote(key_path_02, secret_01)
        same = gatherkeys._keyring_equivalent(key_path_01, key_path_02)
        assert same is True


    def test_different_caps_mixed_quote(self):
        secret_01 = new.generate_auth_key()
        secret_02 = new.generate_auth_key()
        key_path_01 = self.test_dir + "/01.keyring"
        key_path_02 = self.test_dir + "/02.keyring"
        write_key_mon_with_caps_with_tab(key_path_01, secret_01)
        write_key_mon_with_caps_with_tab_quote(key_path_02, secret_02)
        same = gatherkeys._keyring_equivalent(key_path_01, key_path_02)
        assert same is False


    def test_missing_key_1(self):
        secret_02 = new.generate_auth_key()
        key_path_01 = self.test_dir + "/01.keyring"
        key_path_02 = self.test_dir + "/02.keyring"
        write_key_mon_with_caps_with_tab_quote(key_path_02, secret_02)
        with pytest.raises(IOError):
            gatherkeys._keyring_equivalent(key_path_01, key_path_02)


    def test_missing_key_2(self):
        secret_01 = new.generate_auth_key()
        key_path_01 = self.test_dir + "/01.keyring"
        key_path_02 = self.test_dir + "/02.keyring"
        write_key_mon_with_caps_with_tab_quote(key_path_01, secret_01)
        with pytest.raises(IOError):
            gatherkeys._keyring_equivalent(key_path_01, key_path_02)



from ceph_deploy.lib import remoto
from ceph_deploy.hosts.centos.install import repo_install, mirror_install  # noqa
from ceph_deploy.util.paths import gpg
from ceph_deploy.hosts.common import map_components


NON_SPLIT_PACKAGES = ['ceph-osd', 'ceph-mon', 'ceph-mds']


def install(distro, version_kind, version, adjust_repos, **kw):
    packages = map_components(
        NON_SPLIT_PACKAGES,
        kw.pop('components', [])
    )
    gpgcheck = kw.pop('gpgcheck', 1)

    logger = distro.conn.logger
    release = distro.release
    machine = distro.machine_type

    if version_kind in ['stable', 'testing']:
        key = 'release'
    else:
        key = 'autobuild'

    if adjust_repos:
        if distro.packager.name == 'yum':
            distro.packager.install('yum-plugin-priorities')
            # haven't been able to determine necessity of check_obsoletes with DNF
            distro.conn.remote_module.enable_yum_priority_obsoletes()
            logger.warning('check_obsoletes has been enabled for Yum priorities plugin')

        if version_kind in ['stable', 'testing']:
            distro.packager.add_repo_gpg_key(gpg.url(key))

            if version_kind == 'stable':
                url = 'https://download.ceph.com/rpm-{version}/fc{release}/'.format(
                    version=version,
                    release=release,
                    )
            elif version_kind == 'testing':
                url = 'https://download.ceph.com/rpm-testing/fc{release}'.format(
                    release=release,
                    )

            remoto.process.run(
                distro.conn,
                [
                    'rpm',
                    '-Uvh',
                    '--replacepkgs',
                    '--force',
                    '--quiet',
                    '{url}noarch/ceph-release-1-0.fc{release}.noarch.rpm'.format(
                        url=url,
                        release=release,
                        ),
                ]
            )

            # set the right priority
            logger.warning('ensuring that /etc/yum.repos.d/ceph.repo contains a high priority')
            distro.conn.remote_module.set_repo_priority(['Ceph', 'Ceph-noarch', 'ceph-source'])
            logger.warning('altered ceph.repo priorities to contain: priority=1')

        elif version_kind in ['dev', 'dev_commit']:
            logger.info('skipping install of ceph-release package')
            logger.info('repo file will be created manually')
            mirror_install(
                distro,
                'http://gitbuilder.ceph.com/ceph-rpm-fc{release}-{machine}-basic/{sub}/{version}/'.format(
                    release=release.split(".", 1)[0],
                    machine=machine,
                    sub='ref' if version_kind == 'dev' else 'sha1',
                    version=version),
                gpg.url(key),
                adjust_repos=True,
                extra_installs=False,
                gpgcheck=gpgcheck,
            )

        else:
            raise Exception('unrecognized version_kind %s' % version_kind)

    distro.packager.install(
        packages
    )



import pytest

from ceph_deploy.cli import get_parser
from ceph_deploy.tests.util import assert_too_few_arguments


class TestParserRepo(object):

    def setup(self):
        self.parser = get_parser()

    def test_repo_help(self, capsys):
        with pytest.raises(SystemExit):
            self.parser.parse_args('repo --help'.split())
        out, err = capsys.readouterr()
        assert 'usage: ceph-deploy repo' in out
        assert 'positional arguments:' in out
        assert 'optional arguments:' in out

    def test_repo_name_required(self, capsys):
        with pytest.raises(SystemExit):
            self.parser.parse_args('repo'.split())
        out, err = capsys.readouterr()
        assert_too_few_arguments(err)

    def test_repo_host_required(self, capsys):
        with pytest.raises(SystemExit):
            self.parser.parse_args('repo ceph'.split())
        out, err = capsys.readouterr()
        assert_too_few_arguments(err)

    def test_repo_one_host(self):
        args = self.parser.parse_args('repo ceph host1'.split())
        assert args.host == ['host1']

    def test_repo_multiple_hosts(self):
        hostnames = ['host1', 'host2', 'host3']
        args = self.parser.parse_args(['repo', 'ceph'] + hostnames)
        assert frozenset(args.host) == frozenset(hostnames)

    def test_repo_name(self):
        args = self.parser.parse_args('repo ceph host1'.split())
        assert args.repo_name == 'ceph'

    def test_repo_remove_default_is_false(self):
        args = self.parser.parse_args('repo ceph host1'.split())
        assert not args.remove

    def test_repo_remove_set_true(self):
        args = self.parser.parse_args('repo ceph --remove host1'.split())
        assert args.remove

    def test_repo_remove_delete_alias(self):
        args = self.parser.parse_args('repo ceph --delete host1'.split())
        assert args.remove

    def test_repo_url_default_is_none(self):
        args = self.parser.parse_args('repo ceph host1'.split())
        assert args.repo_url is None

    def test_repo_url_custom_path(self):
        args = self.parser.parse_args('repo ceph --repo-url https://ceph.com host1'.split())
        assert args.repo_url == "https://ceph.com"

    def test_repo_gpg_url_default_is_none(self):
        args = self.parser.parse_args('repo ceph host1'.split())
        assert args.gpg_url is None

    def test_repo_gpg_url_custom_path(self):
        args = self.parser.parse_args('repo ceph --gpg-url https://ceph.com/key host1'.split())
        assert args.gpg_url == "https://ceph.com/key"



import logging
from ceph_deploy.lib import remoto
from ceph_deploy.connection import get_local_connection


def can_connect_passwordless(hostname):
    """
    Ensure that current host can SSH remotely to the remote
    host using the ``BatchMode`` option to prevent a password prompt.

    That attempt will error with an exit status of 255 and a ``Permission
    denied`` message or a``Host key verification failed`` message.
    """
    # Ensure we are not doing this for local hosts
    if not remoto.connection.needs_ssh(hostname):
        return True

    logger = logging.getLogger(hostname)
    with get_local_connection(logger) as conn:
        # Check to see if we can login, disabling password prompts
        command = ['ssh', '-CT', '-o', 'BatchMode=yes', hostname]
        out, err, retval = remoto.process.check(conn, command, stop_on_error=False)
        permission_denied_error = b'Permission denied '
        host_key_verify_error = b'Host key verification failed.'
        has_key_error = False
        for line in err:
            if permission_denied_error in line or host_key_verify_error in line:
                has_key_error = True

        if retval == 255 and has_key_error:
            return False
    return True



from mock import Mock

from ceph_deploy import install


class TestSanitizeArgs(object):

    def setup(self):
        self.args = Mock()
        # set the default behavior we set in cli.py
        self.args.default_release = False
        self.args.stable = None

    def test_args_release_not_specified(self):
        self.args.release = None
        result = install.sanitize_args(self.args)
        # XXX
        # we should get `args.release` to be the latest release
        # but we don't want to be updating this test every single
        # time there is a new default value, and we can't programatically
        # change that. Future improvement: make the default release a
        # variable in `ceph_deploy/__init__.py`
        assert result.default_release is True

    def test_args_release_is_specified(self):
        self.args.release = 'dumpling'
        result = install.sanitize_args(self.args)
        assert result.default_release is False

    def test_args_release_stable_is_used(self):
        self.args.stable = 'dumpling'
        result = install.sanitize_args(self.args)
        assert result.release == 'dumpling'

    def test_args_stable_is_not_used(self):
        self.args.release = 'dumpling'
        result = install.sanitize_args(self.args)
        assert result.stable is None


class TestDetectComponents(object):

    def setup(self):
        self.args = Mock()
        # default values for install_* flags
        self.args.install_all = False
        self.args.install_mds = False
        self.args.install_mgr = False
        self.args.install_mon = False
        self.args.install_osd = False
        self.args.install_rgw = False
        self.args.install_tests = False
        self.args.install_common = False
        self.args.repo = False
        self.distro = Mock()

    def test_install_with_repo_option_returns_no_packages(self):
        self.args.repo = True
        result = install.detect_components(self.args, self.distro)
        assert result == []

    def test_install_all_returns_all_packages_deb(self):
        self.args.install_all = True
        self.distro.is_rpm = False
        self.distro.is_deb = True
        result = sorted(install.detect_components(self.args, self.distro))
        assert result == sorted([
            'ceph-osd', 'ceph-mds', 'ceph', 'ceph-mon', 'radosgw'
        ])

    def test_install_all_with_other_options_returns_all_packages_deb(self):
        self.distro.is_rpm = False
        self.distro.is_deb = True
        self.args.install_all = True
        self.args.install_mds = True
        self.args.install_mgr = True
        self.args.install_mon = True
        self.args.install_osd = True
        result = sorted(install.detect_components(self.args, self.distro))
        assert result == sorted([
            'ceph-osd', 'ceph-mds', 'ceph', 'ceph-mon', 'radosgw'
        ])

    def test_install_all_returns_all_packages_rpm(self):
        self.args.install_all = True
        result = sorted(install.detect_components(self.args, self.distro))
        assert result == sorted([
            'ceph-osd', 'ceph-mds', 'ceph', 'ceph-mon', 'ceph-radosgw'
        ])

    def test_install_all_with_other_options_returns_all_packages_rpm(self):
        self.args.install_all = True
        self.args.install_mds = True
        self.args.install_mon = True
        self.args.install_mgr = True
        self.args.install_osd = True
        result = sorted(install.detect_components(self.args, self.distro))
        assert result == sorted([
            'ceph-osd', 'ceph-mds', 'ceph', 'ceph-mon', 'ceph-radosgw'
        ])

    def test_install_only_one_component(self):
        self.args.install_osd = True
        result = install.detect_components(self.args, self.distro)
        assert result == ['ceph-osd']

    def test_install_a_couple_of_components(self):
        self.args.install_osd = True
        self.args.install_mds = True
        self.args.install_mgr = True
        result = sorted(install.detect_components(self.args, self.distro))
        assert result == sorted(['ceph-osd', 'ceph-mds', 'ceph-mgr'])

    def test_install_tests(self):
        self.args.install_tests = True
        result = sorted(install.detect_components(self.args, self.distro))
        assert result == sorted(['ceph-test'])

    def test_install_all_should_be_default_when_no_options_passed(self):
        result = sorted(install.detect_components(self.args, self.distro))
        assert result == sorted([
            'ceph-osd', 'ceph-mds', 'ceph', 'ceph-mon', 'ceph-radosgw'
        ])



def uninstall(distro, purge=False):
    packages = [
        'ceph',
        'ceph-common',
        'libcephfs1',
        'librados2',
        'librbd1',
        'ceph-radosgw',
        ]
    distro.packager.remove(packages)



import socket
import argparse
import re


class RegexMatch(object):
    """
    Performs regular expression match on value.
    If the regular expression pattern matches it will it will return an error
    message that will work with argparse.
    """

    def __init__(self, pattern, statement=None):
        self.string_pattern = pattern
        self.pattern = re.compile(pattern)
        self.statement = statement
        if not self.statement:
            self.statement = "must match pattern %s" % self.string_pattern

    def __call__(self, string):
        match = self.pattern.search(string)
        if match:
            raise argparse.ArgumentError(None, self.statement)
        return string


class Hostname(object):
    """
    Checks wether a given hostname is resolvable in DNS, otherwise raising and
    argparse error.
    """

    def __init__(self, _socket=None):
        self.socket = _socket or socket  # just used for testing

    def __call__(self, string):
        parts = string.split(':', 1)
        name = parts[0]
        host = parts[-1]
        try:
            self.socket.getaddrinfo(host, 0)
        except self.socket.gaierror:
            msg = "hostname: %s is not resolvable" % host
            raise argparse.ArgumentError(None, msg)

        try:
            self.socket.getaddrinfo(name, 0, 0, 0, 0, self.socket.AI_NUMERICHOST)
        except self.socket.gaierror:
            return string  # not an IP
        else:
            msg = '%s must be a hostname not an IP' % name
            raise argparse.ArgumentError(None, msg)

        return string


class Subnet(object):
    """
    A really dumb validator to ensure that we are receiving a subnet (or
    something that actually looks like a subnet).

    It doesn't enforce at all the constraints of proper validation as that has
    its own set of caveats that are difficult to implement given that
    ceph-deploy doesn't (should not) include third party dependencies.
    """

    def __call__(self, string):
        ip = string.split('/')[0]
        ip_parts = ip.split('.')

        if len(ip_parts) != 4:
            err = "subnet must have at least 4 numbers separated by dots like x.x.x.x/xx, but got: %s" % string
            raise argparse.ArgumentError(None, err)

        if [i for i in ip_parts[:4] if i.isalpha()]:  # only numbers
            err = "subnet must have digits separated by dots like x.x.x.x/xx, but got: %s" % string
            raise argparse.ArgumentError(None, err)

        if len(string.split('/')) != 2:
            err = "subnet must contain a slash, like x.x.x.x/xx, but got: %s" % string
            raise argparse.ArgumentError(None, err)

        return string



import pytest

from ceph_deploy.cli import get_parser
from ceph_deploy.tests.util import assert_too_few_arguments


class TestParserCalamari(object):

    def setup(self):
        self.parser = get_parser()

    def test_calamari_help(self, capsys):
        with pytest.raises(SystemExit):
            self.parser.parse_args('calamari --help'.split())
        out, err = capsys.readouterr()
        assert 'usage: ceph-deploy calamari' in out
        assert 'positional arguments:' in out
        assert 'optional arguments:' in out

    def test_calamari_connect_help(self, capsys):
        with pytest.raises(SystemExit):
            self.parser.parse_args('calamari connect --help'.split())
        out, err = capsys.readouterr()
        assert 'usage: ceph-deploy calamari connect' in out
        assert 'positional arguments:' in out
        assert 'optional arguments:' in out

    def test_calamari_connect_host_required(self, capsys):
        with pytest.raises(SystemExit):
            self.parser.parse_args('calamari connect'.split())
        out, err = capsys.readouterr()
        assert_too_few_arguments(err)

    def test_calamari_connect_one_host(self):
        args = self.parser.parse_args('calamari connect host1'.split())
        assert args.hosts == ['host1']

    def test_calamari_connect_multiple_hosts(self):
        hostnames = ['host1', 'host2', 'host3']
        args = self.parser.parse_args('calamari connect'.split() + hostnames)
        assert args.hosts == hostnames

    def test_calamari_connect_master_default_is_none(self):
        args = self.parser.parse_args('calamari connect host1'.split())
        assert args.master is None

    def test_calamari_connect_master_custom(self):
        args = self.parser.parse_args('calamari connect --master master.ceph.com host1'.split())
        assert args.master == "master.ceph.com"



import pytest

from ceph_deploy.cli import get_parser

SUBCMDS_WITH_ARGS = ['add', 'destroy', 'create']
SUBCMDS_WITHOUT_ARGS = ['create', 'create-initial']


class TestParserMON(object):

    def setup(self):
        self.parser = get_parser()

    def test_mon_help(self, capsys):
        with pytest.raises(SystemExit):
            self.parser.parse_args('mon --help'.split())
        out, err = capsys.readouterr()
        assert 'usage: ceph-deploy mon' in out
        assert 'positional arguments:' in out
        assert 'optional arguments:' in out

    @pytest.mark.parametrize('cmd', SUBCMDS_WITH_ARGS)
    def test_mon_valid_subcommands_with_args(self, cmd, capsys):
        args = self.parser.parse_args(['mon'] + ['%s' % cmd] + ['host1'])
        assert args.subcommand == cmd

    @pytest.mark.parametrize('cmd', SUBCMDS_WITHOUT_ARGS)
    def test_mon_valid_subcommands_without_args(self, cmd, capsys):
        args = self.parser.parse_args(['mon'] + ['%s' % cmd])
        assert args.subcommand == cmd

    def test_mon_invalid_subcommand(self, capsys):
        with pytest.raises(SystemExit):
            self.parser.parse_args('mon bork'.split())
        out, err = capsys.readouterr()
        assert 'invalid choice' in err

    def test_mon_create_initial_help(self, capsys):
        with pytest.raises(SystemExit):
            self.parser.parse_args('mon create-initial --help'.split())
        out, err = capsys.readouterr()
        assert 'usage: ceph-deploy mon create-initial' in out

    def test_mon_create_initial_keyrings_default_none(self):
        args = self.parser.parse_args('mon create-initial'.split())
        assert args.keyrings is None

    def test_mon_create_initial_keyrings_custom_dir(self):
        args = self.parser.parse_args('mon create-initial --keyrings /tmp/keys'.split())
        assert args.keyrings == "/tmp/keys"

    def test_mon_create_initial_keyrings_host_raises_err(self):
        with pytest.raises(SystemExit):
            self.parser.parse_args('mon create-initial test1'.split())

    def test_mon_create_help(self, capsys):
        with pytest.raises(SystemExit):
            self.parser.parse_args('mon create --help'.split())
        out, err = capsys.readouterr()
        assert 'usage: ceph-deploy mon create' in out

    def test_mon_create_keyrings_default_none(self):
        args = self.parser.parse_args('mon create'.split())
        assert args.keyrings is None

    def test_mon_create_keyrings_custom_dir(self):
        args = self.parser.parse_args('mon create --keyrings /tmp/keys'.split())
        assert args.keyrings == "/tmp/keys"

    def test_mon_create_single_host(self):
        args = self.parser.parse_args('mon create test1'.split())
        assert args.mon == ['test1']

    def test_mon_create_multi_host(self):
        hosts = ['host1', 'host2', 'host3']
        args = self.parser.parse_args('mon create'.split() + hosts)
        assert args.mon == hosts

    def test_mon_add_help(self, capsys):
        with pytest.raises(SystemExit):
            self.parser.parse_args('mon add --help'.split())
        out, err = capsys.readouterr()
        assert 'usage: ceph-deploy mon add' in out

    def test_mon_add_address_default_none(self):
        args = self.parser.parse_args('mon add test1'.split())
        assert args.address is None

    def test_mon_add_address_custom_addr(self):
        args = self.parser.parse_args('mon add test1 --address 10.10.0.1'.split())
        assert args.address == '10.10.0.1'

    def test_mon_add_no_host_raises_err(self):
        with pytest.raises(SystemExit):
            self.parser.parse_args('mon add'.split())

    def test_mon_add_one_host_okay(self):
        args = self.parser.parse_args('mon add test1'.split())
        assert args.mon == ["test1"]

    def test_mon_add_multi_host_raises_err(self):
        with pytest.raises(SystemExit):
            self.parser.parse_args('mon add test1 test2'.split())

    def test_mon_destroy_help(self, capsys):
        with pytest.raises(SystemExit):
            self.parser.parse_args('mon destroy --help'.split())
        out, err = capsys.readouterr()
        assert 'usage: ceph-deploy mon destroy' in out

    def test_mon_destroy_no_host_raises_err(self):
        with pytest.raises(SystemExit):
            self.parser.parse_args('mon destroy'.split())

    def test_mon_destroy_one_host_okay(self):
        args = self.parser.parse_args('mon destroy test1'.split())
        assert args.mon == ["test1"]

    def test_mon_destroy_multi_host(self):
        hosts = ['host1', 'host2', 'host3']
        args = self.parser.parse_args('mon destroy'.split() + hosts)
        assert args.mon == hosts



import errno
import logging
import os

from ceph_deploy import conf
from ceph_deploy import exc
from ceph_deploy import hosts
from ceph_deploy.util import system
from ceph_deploy.lib import remoto
from ceph_deploy.cliutil import priority


LOG = logging.getLogger(__name__)


def get_bootstrap_rgw_key(cluster):
    """
    Read the bootstrap-rgw key for `cluster`.
    """
    path = '{cluster}.bootstrap-rgw.keyring'.format(cluster=cluster)
    try:
        with open(path, 'rb') as f:
            return f.read()
    except IOError:
        raise RuntimeError('bootstrap-rgw keyring not found; run \'gatherkeys\'')


def create_rgw(distro, name, cluster, init):
    conn = distro.conn

    path = '/var/lib/ceph/radosgw/{cluster}-{name}'.format(
        cluster=cluster,
        name=name
        )

    conn.remote_module.safe_makedirs(path)

    bootstrap_keyring = '/var/lib/ceph/bootstrap-rgw/{cluster}.keyring'.format(
        cluster=cluster
        )

    keypath = os.path.join(path, 'keyring')

    stdout, stderr, returncode = remoto.process.check(
        conn,
        [
            'ceph',
            '--cluster', cluster,
            '--name', 'client.bootstrap-rgw',
            '--keyring', bootstrap_keyring,
            'auth', 'get-or-create', 'client.{name}'.format(name=name),
            'osd', 'allow rwx',
            'mon', 'allow rw',
            '-o',
            os.path.join(keypath),
        ]
    )
    if returncode > 0 and returncode != errno.EACCES:
        for line in stderr:
            conn.logger.error(line)
        for line in stdout:
            # yes stdout as err because this is an error
            conn.logger.error(line)
        conn.logger.error('exit code from command was: %s' % returncode)
        raise RuntimeError('could not create rgw')

        remoto.process.check(
            conn,
            [
                'ceph',
                '--cluster', cluster,
                '--name', 'client.bootstrap-rgw',
                '--keyring', bootstrap_keyring,
                'auth', 'get-or-create', 'client.{name}'.format(name=name),
                'osd', 'allow *',
                'mon', 'allow *',
                '-o',
                os.path.join(keypath),
            ]
        )

    conn.remote_module.touch_file(os.path.join(path, 'done'))
    conn.remote_module.touch_file(os.path.join(path, init))

    if init == 'upstart':
        remoto.process.run(
            conn,
            [
                'initctl',
                'emit',
                'radosgw',
                'cluster={cluster}'.format(cluster=cluster),
                'id={name}'.format(name=name),
            ],
            timeout=7
        )
    elif init == 'sysvinit':
        remoto.process.run(
            conn,
            [
                'service',
                'ceph-radosgw',
                'start',
            ],
            timeout=7
        )
        if distro.is_el:
            system.enable_service(distro.conn, service='ceph-radosgw')
    elif init == 'systemd':
        remoto.process.run(
            conn,
            [
                'systemctl',
                'enable',
                'ceph-radosgw@{name}'.format(name=name),
            ],
            timeout=7
        )
        remoto.process.run(
            conn,
            [
                'systemctl',
                'start',
                'ceph-radosgw@{name}'.format(name=name),
            ],
            timeout=7
        )
        remoto.process.run(
            conn,
            [
                'systemctl',
                'enable',
                'ceph.target',
            ],
            timeout=7
        )


def rgw_create(args):
    conf_data = conf.ceph.load_raw(args)
    LOG.debug(
        'Deploying rgw, cluster %s hosts %s',
        args.cluster,
        ' '.join(':'.join(x or '' for x in t) for t in args.rgw),
        )

    key = get_bootstrap_rgw_key(cluster=args.cluster)

    bootstrapped = set()
    errors = 0
    for hostname, name in args.rgw:
        try:
            distro = hosts.get(hostname, username=args.username)
            rlogger = distro.conn.logger
            LOG.info(
                'Distro info: %s %s %s',
                distro.name,
                distro.release,
                distro.codename
            )
            LOG.debug('remote host will use %s', distro.init)

            if hostname not in bootstrapped:
                bootstrapped.add(hostname)
                LOG.debug('deploying rgw bootstrap to %s', hostname)
                distro.conn.remote_module.write_conf(
                    args.cluster,
                    conf_data,
                    args.overwrite_conf,
                )

                path = '/var/lib/ceph/bootstrap-rgw/{cluster}.keyring'.format(
                    cluster=args.cluster,
                )

                if not distro.conn.remote_module.path_exists(path):
                    rlogger.warning('rgw keyring does not exist yet, creating one')
                    distro.conn.remote_module.write_keyring(path, key)

            create_rgw(distro, name, args.cluster, distro.init)
            distro.conn.exit()
            LOG.info(
                ('The Ceph Object Gateway (RGW) is now running on host %s and '
                 'default port %s'),
                hostname,
                '7480'
            )
        except RuntimeError as e:
            LOG.error(e)
            errors += 1

    if errors:
        raise exc.GenericError('Failed to create %d RGWs' % errors)


def rgw(args):
    if args.subcommand == 'create':
        rgw_create(args)
    else:
        LOG.error('subcommand %s not implemented', args.subcommand)


def colon_separated(s):
    host = s
    name = s
    if s.count(':') == 1:
        (host, name) = s.split(':')
    name = 'rgw.' + name
    return (host, name)


@priority(30)
def make(parser):
    """
    Ceph RGW daemon management
    """
    rgw_parser = parser.add_subparsers(dest='subcommand')
    rgw_parser.required = True
    rgw_create = rgw_parser.add_parser(
        'create',
        help='Create an RGW instance'
        )
    rgw_create.add_argument(
        'rgw',
        metavar='HOST[:NAME]',
        nargs='+',
        type=colon_separated,
        help='host (and optionally the daemon name) to deploy on. \
                NAME is automatically prefixed with \'rgw.\'',
        )
    parser.set_defaults(
        func=rgw,
        )



class DeployError(Exception):
    """
    Unknown deploy error
    """

    def __str__(self):
        doc = self.__doc__.strip()
        return ': '.join([doc] + [str(a) for a in self.args])


class UnableToResolveError(DeployError):
    """
    Unable to resolve host
    """


class ClusterExistsError(DeployError):
    """
    Cluster config exists already
    """


class ConfigError(DeployError):
    """
    Cannot load config
    """


class NeedHostError(DeployError):
    """
    No hosts specified to deploy to.
    """


class NeedMonError(DeployError):
    """
    Cannot find nodes with ceph-mon.
    """


class NeedDiskError(DeployError):
    """
    Must supply disk/path argument
    """


class UnsupportedPlatform(DeployError):
    """
    Platform is not supported
    """
    def __init__(self, distro, codename, release):
        self.distro = distro
        self.codename = codename
        self.release = release

    def __str__(self):
        return '{doc}: {distro} {codename} {release}'.format(
            doc=self.__doc__.strip(),
            distro=self.distro,
            codename=self.codename,
            release=self.release,
        )


class ExecutableNotFound(DeployError):
    """
    Could not locate executable
    """
    def __init__(self, executable, host):
        self.executable = executable
        self.host = host

    def __str__(self):
        return "{doc} '{executable}' make sure it is installed and available on {host}".format(
            doc=self.__doc__.strip(),
            executable=self.executable,
            host=self.host,
        )


class MissingPackageError(DeployError):
    """
    A required package or command is missing
    """
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return self.message


class GenericError(DeployError):
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return self.message


class ClusterNameError(DeployError):
    """
    Problem encountered with custom cluster name
    """
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return self.message


class KeyNotFoundError(DeployError):
    """
    Could not find keyring file
    """
    def __init__(self, keyring, hosts):
        self.keyring = keyring
        self.hosts = hosts

    def __str__(self):
        return '{doc}: {keys}'.format(
            doc=self.__doc__.strip(),
            keys=', '.join(
                [self.keyring.format(hostname=host) +
                 " on host {hostname}".format(hostname=host)
                 for host in self.hosts]
            )
        )



from ceph_deploy import gatherkeys
from ceph_deploy import new
import mock
import json
import copy


remoto_process_check_success_output = {
        "name": "ceph-node1",
        "rank": 0,
        "state": "leader",
        "election_epoch": 6,
        "quorum": [
            0,
            1,
            2
        ],
        "outside_quorum": [],
        "extra_probe_peers": [
            "192.168.99.125:6789\/0",
            "192.168.99.126:6789\/0"
        ],
        "sync_provider": [],
        "monmap": {
            "epoch": 1,
            "fsid": "4dbee7eb-929b-4f3f-ad23-8a4e47235e40",
            "modified": "2016-04-11 05:35:21.665220",
            "created": "2016-04-11 05:35:21.665220",
            "mons": [
                {
                    "rank": 0,
                    "name": "host0",
                    "addr": "192.168.99.124:6789\/0"
                },
                {
                    "rank": 1,
                    "name": "host1",
                    "addr": "192.168.99.125:6789\/0"
                },
                {
                    "rank": 2,
                    "name": "host2",
                    "addr": "192.168.99.126:6789\/0"
                }
            ]
        }
    }


class mock_remote_module(object):
    def get_file(self, path):
        return self.get_file_result

    def shortname(self):
        hostname_split = self.longhostname.split('.')
        return hostname_split[0]

class mock_conn(object):
    def __init__(self):
        self.remote_module = mock_remote_module()


class mock_distro(object):
    def __init__(self):
        self.conn = mock_conn()
        

def mock_hosts_get_file_key_content(host, **kwargs):
    output = mock_distro()
    mon_keyring = '[mon.]\nkey = %s\ncaps mon = allow *\n' % new.generate_auth_key()
    output.conn.remote_module.get_file_result = mon_keyring.encode('utf-8')
    output.conn.remote_module.longhostname = host
    return output


def mock_hosts_get_file_key_content_none(host, **kwargs):
    output = mock_distro()
    output.conn.remote_module.get_file_result = None
    output.conn.remote_module.longhostname = host
    return output


def mock_gatherkeys_missing_success(args, distro, rlogger, path_keytype_mon, keytype, dest_dir):
    return True


def mock_gatherkeys_missing_fail(args, distro, rlogger, path_keytype_mon, keytype, dest_dir):
    return False


def mock_remoto_process_check_success(conn, args):
    out = json.dumps(remoto_process_check_success_output,sort_keys=True, indent=4)
    return out.encode('utf-8').split(b'\n'), [], 0


def mock_remoto_process_check_rc_error(conn, args):
    return [b""], [b"this failed\n"], 1


def mock_remoto_process_check_out_not_json(conn, args):
    return [b"}bad output{"], [b""], 0


def mock_remoto_process_check_out_missing_quorum(conn, args):
    outdata = copy.deepcopy(remoto_process_check_success_output)
    del outdata["quorum"]
    out = json.dumps(outdata,sort_keys=True, indent=4)
    return out.encode('utf-8').split(b'\n'), [], 0


def mock_remoto_process_check_out_missing_quorum_1(conn, args):
    outdata = copy.deepcopy(remoto_process_check_success_output)
    del outdata["quorum"][1]
    out = json.dumps(outdata,sort_keys=True, indent=4)
    return out.encode('utf-8').split(b'\n'), [], 0


def mock_remoto_process_check_out_missing_monmap(conn, args):
    outdata = copy.deepcopy(remoto_process_check_success_output)
    del outdata["monmap"]
    out = json.dumps(outdata,sort_keys=True, indent=4)
    return out.encode('utf-8').split(b'\n'), [], 0


def mock_remoto_process_check_out_missing_mons(conn, args):
    outdata = copy.deepcopy(remoto_process_check_success_output)
    del outdata["monmap"]["mons"]
    out = json.dumps(outdata,sort_keys=True, indent=4)
    return out.encode('utf-8').split(b'\n'), [], 0


def mock_remoto_process_check_out_missing_monmap_host1(conn, args):
    outdata = copy.deepcopy(remoto_process_check_success_output)
    del outdata["monmap"]["mons"][1]
    out = json.dumps(outdata,sort_keys=True, indent=4)
    return out.encode('utf-8').split(b'\n'), [], 0


class TestGatherKeysWithMon(object):
    """
    Test gatherkeys_with_mon function
    """
    def setup(self):
        self.args = mock.Mock()
        self.args.cluster = "ceph"
        self.args.mon = ['host1']
        self.host = 'host1'
        self.test_dir = '/tmp'


    @mock.patch('ceph_deploy.gatherkeys.gatherkeys_missing', mock_gatherkeys_missing_success)
    @mock.patch('ceph_deploy.lib.remoto.process.check', mock_remoto_process_check_success)
    @mock.patch('ceph_deploy.hosts.get', mock_hosts_get_file_key_content)
    def test_success(self):
        rc = gatherkeys.gatherkeys_with_mon(self.args, self.host, self.test_dir)
        assert rc is True


    @mock.patch('ceph_deploy.gatherkeys.gatherkeys_missing', mock_gatherkeys_missing_success)
    @mock.patch('ceph_deploy.lib.remoto.process.check', mock_remoto_process_check_success)
    @mock.patch('ceph_deploy.hosts.get', mock_hosts_get_file_key_content_none)
    def test_monkey_none(self):
        rc = gatherkeys.gatherkeys_with_mon(self.args, self.host, self.test_dir)
        assert rc is False


    @mock.patch('ceph_deploy.gatherkeys.gatherkeys_missing', mock_gatherkeys_missing_fail)
    @mock.patch('ceph_deploy.lib.remoto.process.check', mock_remoto_process_check_success)
    @mock.patch('ceph_deploy.hosts.get', mock_hosts_get_file_key_content)
    def test_missing_fail(self):
        rc = gatherkeys.gatherkeys_with_mon(self.args, self.host, self.test_dir)
        assert rc is False


    @mock.patch('ceph_deploy.gatherkeys.gatherkeys_missing', mock_gatherkeys_missing_success)
    @mock.patch('ceph_deploy.lib.remoto.process.check', mock_remoto_process_check_rc_error)
    @mock.patch('ceph_deploy.hosts.get', mock_hosts_get_file_key_content)
    def test_remoto_process_check_rc_error(self):
        rc = gatherkeys.gatherkeys_with_mon(self.args, self.host, self.test_dir)
        assert rc is False


    @mock.patch('ceph_deploy.gatherkeys.gatherkeys_missing', mock_gatherkeys_missing_success)
    @mock.patch('ceph_deploy.lib.remoto.process.check', mock_remoto_process_check_out_not_json)
    @mock.patch('ceph_deploy.hosts.get', mock_hosts_get_file_key_content)
    def test_remoto_process_check_out_not_json(self):
        rc = gatherkeys.gatherkeys_with_mon(self.args, self.host, self.test_dir)
        assert rc is False

    @mock.patch('ceph_deploy.gatherkeys.gatherkeys_missing', mock_gatherkeys_missing_success)
    @mock.patch('ceph_deploy.lib.remoto.process.check', mock_remoto_process_check_out_missing_quorum)
    @mock.patch('ceph_deploy.hosts.get', mock_hosts_get_file_key_content)
    def test_remoto_process_check_out_missing_quorum(self):
        rc = gatherkeys.gatherkeys_with_mon(self.args, self.host, self.test_dir)
        assert rc is False


    @mock.patch('ceph_deploy.gatherkeys.gatherkeys_missing', mock_gatherkeys_missing_success)
    @mock.patch('ceph_deploy.lib.remoto.process.check', mock_remoto_process_check_out_missing_quorum_1)
    @mock.patch('ceph_deploy.hosts.get', mock_hosts_get_file_key_content)
    def test_remoto_process_check_out_missing_quorum_1(self):
        rc = gatherkeys.gatherkeys_with_mon(self.args, self.host, self.test_dir)
        assert rc is False


    @mock.patch('ceph_deploy.gatherkeys.gatherkeys_missing', mock_gatherkeys_missing_success)
    @mock.patch('ceph_deploy.lib.remoto.process.check', mock_remoto_process_check_out_missing_mons)
    @mock.patch('ceph_deploy.hosts.get', mock_hosts_get_file_key_content)
    def test_remoto_process_check_out_missing_mon(self):
        rc = gatherkeys.gatherkeys_with_mon(self.args, self.host, self.test_dir)
        assert rc is False


    @mock.patch('ceph_deploy.gatherkeys.gatherkeys_missing', mock_gatherkeys_missing_success)
    @mock.patch('ceph_deploy.lib.remoto.process.check', mock_remoto_process_check_out_missing_monmap_host1)
    @mock.patch('ceph_deploy.hosts.get', mock_hosts_get_file_key_content)
    def test_remoto_process_check_out_missing_monmap_host1(self):
        rc = gatherkeys.gatherkeys_with_mon(self.args, self.host, self.test_dir)
        assert rc is False



from os.path import join
from collections import namedtuple

# Base Path for ceph
base_path = '/var/lib/ceph'

# Base run Path
base_run_path = '/var/run/ceph'

tmp_path = join(base_path, 'tmp')

mon_path = join(base_path, 'mon')

mgr_path = join(base_path, 'mgr')

mds_path = join(base_path, 'mds')

osd_path = join(base_path, 'osd')

# Default package components to install
_base_components = [
    'ceph',
    'ceph-osd',
    'ceph-mds',
    'ceph-mon',
]

default_components = namedtuple('DefaultComponents', ['rpm', 'deb'])

# the difference here is because RPMs currently name the radosgw differently than DEBs.
# TODO: This needs to get unified once the packaging naming gets consistent
default_components.rpm = tuple(_base_components + ['ceph-radosgw'])
default_components.deb = tuple(_base_components + ['radosgw'])

gpg_key_base_url = "download.ceph.com/keys/"



import logging
import os
import subprocess
import sys


LOG = logging.getLogger(__name__)


def _prepend_path(env):
    """
    Make sure the PATH contains the location where the Python binary
    lives. This makes sure cli tools installed in a virtualenv work.
    """
    if env is None:
        env = os.environ
    env = dict(env)
    new = os.path.dirname(sys.executable)
    path = env.get('PATH')
    if path is not None:
        new = new + ':' + path
    env['PATH'] = new
    return env


class CLIFailed(Exception):
    """CLI tool failed"""

    def __init__(self, args, status):
        self.args = args
        self.status = status

    def __str__(self):
        return '{doc}: {args}: exited with status {status}'.format(
            doc=self.__doc__,
            args=self.args,
            status=self.status,
            )


class CLIProcess(object):
    def __init__(self, **kw):
        self.kw = kw

    def __enter__(self):
        try:
            self.p = subprocess.Popen(**self.kw)
        except OSError as e:
            raise AssertionError(
                'CLI tool {args!r} does not work: {err}'.format(
                    args=self.kw['args'],
                    err=e,
                    ),
                )
        else:
            return self.p

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.p.wait()
        if self.p.returncode != 0:
            err = CLIFailed(
                args=self.kw['args'],
                status=self.p.returncode,
                )
            if exc_type is None:
                # nothing else raised, so we should complain; if
                # something else failed, we'll just log
                raise err
            else:
                LOG.error(str(err))


class CLITester(object):
    # provide easy way for caller to access the exception class
    # without importing us
    Failed = CLIFailed

    def __init__(self, tmpdir):
        self.tmpdir = tmpdir

    def __call__(self, **kw):
        kw.setdefault('cwd', str(self.tmpdir))
        kw['env'] = _prepend_path(kw.get('env'))
        kw['env']['COLUMNS'] = '80'
        return CLIProcess(**kw)


def pytest_funcarg__cli(request):
    """
    Test command line behavior.
    """

    # the tmpdir here will be the same value as the test function
    # sees; we rely on that to let caller prepare and introspect
    # any files the cli tool will read or create
    tmpdir = request.getfuncargvalue('tmpdir')

    return CLITester(tmpdir=tmpdir)



from ceph_deploy.exc import ExecutableNotFound
from ceph_deploy.util import system, versions
from ceph_deploy.lib import remoto


class Ceph(object):
    """
    Determine different aspects of the Ceph package, like ``version`` and path
    ``executable``. Although mostly provide a version object that helps for
    parsing and comparing.
    """

    def __init__(self, conn, _check=None):
        self.conn = conn
        self._check = _check or remoto.process.check

    @property
    def installed(self):
        """
        If the ``ceph`` executable exists, then Ceph is installed. Should
        probably be revisited if different components do not have the ``ceph``
        executable (this is currently provided by ``ceph-common``).
        """
        return bool(self.executable)

    @property
    def executable(self):
        try:
            return system.executable_path(self.conn, 'ceph')
        except ExecutableNotFound:
            return None

    def _get_version_output(self):
        """
        Ignoring errors, call `ceph --version` and return only the version
        portion of the output. For example, output like::

            ceph version 9.0.1-1234kjd (asdflkj2k3jh234jhg)

        Would return::

            9.0.1-1234kjd
        """
        if not self.executable:
            return ''
        command = [self.executable, '--version']
        out, _, _ = self._check(self.conn, command)
        try:
            return out.decode('utf-8').split()[2]
        except IndexError:
            return ''

    @property
    def version(self):
        """
        Return a version object (see
        :mod:``ceph_deploy.util.versions.NormalizedVersion``)
        """
        return versions.parse_version(self._get_version_output)


# callback helpers

def ceph_is_installed(module):
    """
    A helper callback to be executed after the connection is made to ensure
    that Ceph is installed.
    """
    ceph_package = Ceph(module.conn)
    if not ceph_package.installed:
        host = module.conn.hostname
        raise RuntimeError(
            'ceph needs to be installed in remote host: %s' % host
        )



import logging
import errno

from .cliutil import priority


LOG = logging.getLogger(__name__)


def forgetkeys(args):
    import os
    for f in [
        'mon',
        'client.admin',
        'bootstrap-osd',
        'bootstrap-mds',
        'bootstrap-rgw',
        ]:
        try:
            os.unlink('{cluster}.{what}.keyring'.format(
                    cluster=args.cluster,
                    what=f,
                    ))
        except OSError as e:
            if e.errno == errno.ENOENT:
                pass
            else:
                raise

@priority(100)
def make(parser):
    """
    Remove authentication keys from the local directory.
    """
    parser.set_defaults(
        func=forgetkeys,
        )



import contextlib
import os


@contextlib.contextmanager
def directory(path):
    prev = os.open('.', os.O_RDONLY | os.O_DIRECTORY)
    try:
        os.chdir(path)
        yield
    finally:
        os.fchdir(prev)
        os.close(prev)



"""
Comosd paths for osd, based on the constant file paths defined in
``ceph_deploy.util.constants``.
All functions return a string representation of the absolute path
construction.
"""
from os.path import join
from ceph_deploy.util import constants


def base(cluster):
    cluster = "%s-" % cluster
    return join(constants.osd_path, cluster)



import traceback
from .log import reporting
from .util import admin_command, RemoteError


def _remote_run(channel, cmd, **kw):
    import subprocess
    import sys
    from select import select
    stop_on_nonzero = kw.pop('stop_on_nonzero', True)

    process = subprocess.Popen(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        close_fds=True,
        **kw
    )

    while True:
        reads, _, _ = select(
            [process.stdout.fileno(), process.stderr.fileno()],
            [], []
        )

        for descriptor in reads:
            if descriptor == process.stdout.fileno():
                read = process.stdout.readline()
                if read:
                    channel.send({'debug': read})
                    sys.stdout.flush()

            if descriptor == process.stderr.fileno():
                read = process.stderr.readline()
                if read:
                    channel.send({'warning': read})
                    sys.stderr.flush()

        if process.poll() is not None:
            # ensure we do not have anything pending in stdout or stderr
            # unfortunately, we cannot abstract this repetitive loop into its
            # own function because execnet does not allow for non-global (or
            # even nested functions). This must be repeated here.
            while True:
                err_read = out_read = None
                for descriptor in reads:
                    if descriptor == process.stdout.fileno():
                        out_read = process.stdout.readline()
                        if out_read:
                            channel.send({'debug': out_read})
                            sys.stdout.flush()

                    if descriptor == process.stderr.fileno():
                        err_read = process.stderr.readline()
                        if err_read:
                            channel.send({'warning': err_read})
                            sys.stderr.flush()
                # At this point we have gone through all the possible
                # descriptors and `read` was empty, so we now can break out of
                # this since all stdout/stderr has been properly flushed to
                # logging
                if not err_read and not out_read:
                    break

            break

    returncode = process.wait()
    if returncode != 0:
        if stop_on_nonzero:
            raise RuntimeError(
                "command returned non-zero exit status: %s" % returncode
            )
        else:
            channel.send({'warning': "command returned non-zero exit status: %s" % returncode})


def extend_path(conn, arguments):
    """
    get the remote environment's env so we can explicitly add the path without
    wiping out everything
    """
    # retrieve the remote environment variables for the host
    try:
        result = conn.gateway.remote_exec("import os; channel.send(os.environ.copy())")
        env = result.receive()
    except Exception:
        conn.logger.exception('failed to retrieve the remote environment variables')
        env = {}

    # get the $PATH and extend it (do not overwrite)
    path = env.get('PATH', '')
    env['PATH'] = path + '/usr/local/bin:/bin:/usr/bin:/usr/local/sbin:/usr/sbin:/sbin'
    arguments['env'] = env

    return arguments


def run(conn, command, exit=False, timeout=None, **kw):
    """
    A real-time-logging implementation of a remote subprocess.Popen call where
    a command is just executed on the remote end and no other handling is done.

    :param conn: A connection oject
    :param command: The command to pass in to the remote subprocess.Popen
    :param exit: If this call should close the connection at the end
    :param timeout: How many seconds to wait after no remote data is received
                    (defaults to wait for ever)
    """
    stop_on_error = kw.pop('stop_on_error', True)
    if not kw.get('env'):
        # get the remote environment's env so we can explicitly add
        # the path without wiping out everything
        kw = extend_path(conn, kw)

    timeout = timeout or conn.global_timeout
    conn.logger.info('Running command: %s' % ' '.join(admin_command(conn.sudo, command)))
    result = conn.execute(_remote_run, cmd=command, **kw)
    try:
        reporting(conn, result, timeout)
    except Exception:
        remote_trace = traceback.format_exc()
        remote_error = RemoteError(remote_trace)
        if remote_error.exception_name == 'RuntimeError':
            conn.logger.error(remote_error.exception_line)
        else:
            for tb_line in remote_trace.split('\n'):
                conn.logger.error(tb_line)
        if stop_on_error:
            raise RuntimeError(
                'Failed to execute command: %s' % ' '.join(command)
            )
    if exit:
        conn.exit()


def _remote_check(channel, cmd, **kw):
    import subprocess

    process = subprocess.Popen(
        cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, **kw
    )
    stdout = process.stdout.read().splitlines()
    stderr = process.stderr.read().splitlines()
    channel.send((stdout, stderr, process.wait()))


def check(conn, command, exit=False, timeout=None, **kw):
    """
    Execute a remote command with ``subprocess.Popen`` but report back the
    results in a tuple with three items: stdout, stderr, and exit status.

    This helper function *does not* provide any logging as it is the caller's
    responsibility to do so.
    """
    stop_on_error = kw.pop('stop_on_error', True)
    timeout = timeout or conn.global_timeout
    if not kw.get('env'):
        # get the remote environment's env so we can explicitly add
        # the path without wiping out everything
        kw = extend_path(conn, kw)

    conn.logger.info('Running command: %s' % ' '.join(admin_command(conn.sudo, command)))
    result = conn.execute(_remote_check, cmd=command, **kw)
    response = None
    try:
        response = result.receive(timeout)
    except Exception as err:
        # the things we need to do here :(
        # because execnet magic, we cannot catch this as
        # `except TimeoutError`
        if err.__class__.__name__ == 'TimeoutError':
            msg = 'No data was received after %s seconds, disconnecting...' % timeout
            conn.logger.warning(msg)
            # there is no stdout, stderr, or exit code but make the exit code
            # an error condition (non-zero) regardless
            return [], [], -1
        else:
            remote_trace = traceback.format_exc()
            remote_error = RemoteError(remote_trace)
            if remote_error.exception_name == 'RuntimeError':
                conn.logger.error(remote_error.exception_line)
            else:
                for tb_line in remote_trace.split('\n'):
                    conn.logger.error(tb_line)
            if stop_on_error:
                raise RuntimeError(
                    'Failed to execute command: %s' % ' '.join(command)
                )
    if exit:
        conn.exit()
    return response




from __future__ import print_function

import subprocess
import os
from os import path
import re
import traceback
import sys


error_msg = """
This library depends on sources fetched when packaging that failed to be
retrieved.

This means that it will *not* work as expected. Errors encountered:
"""


def run(cmd):
    print('[vendoring] Running command: %s' % ' '.join(cmd))
    try:
        result = subprocess.Popen(
            cmd,
            stderr=subprocess.PIPE,
            stdout=subprocess.PIPE
        )
    except Exception:
        # if building with python2.5 this makes it compatible
        _, error, _ = sys.exc_info()
        print_error([], traceback.format_exc(error).split('\n'))
        raise SystemExit(1)

    if result.wait():
        print_error(result.stdout.readlines(), result.stderr.readlines())

    return result.returncode


def print_error(stdout, stderr):
    print('*'*80)
    print(error_msg)
    for line in stdout:
        print(line)
    for line in stderr:
        print(line)
    print('*'*80)


def vendor_library(name, version, cmd=None):
    this_dir = path.dirname(path.abspath(__file__))
    vendor_dest = path.join(this_dir, 'ceph_deploy/lib/vendor/%s' % name)
    vendor_init = path.join(vendor_dest, '__init__.py')
    vendor_src = path.join(this_dir, name)
    vendor_module = path.join(vendor_src, name)
    current_dir = os.getcwd()

    if path.exists(vendor_src):
        run(['rm', '-rf', vendor_src])

    if path.exists(vendor_init):
        # The following read/regex is done so that we can parse module metadata without the need
        # to import it. Module metadata is defined as variables with double underscores. We are
        # particularly insteresting in the version string, so we look into single or double quoted
        # values, like:  __version__ = '1.0'
        module_file = open(vendor_init).read()
        metadata = dict(re.findall(r"__([a-z]+)__\s*=\s*['\"]([^'\"]*)['\"]", module_file))
        if metadata.get('version') != version:
            run(['rm', '-rf', vendor_dest])

    if not path.exists(vendor_dest):
        rc = run(['git', 'clone', 'git://git.ceph.com/%s' % name])
        if rc:
            print("%s: git clone failed using ceph.com url with rc %s, trying github.com" % (path.basename(__file__), rc))
            run(['git', 'clone', 'https://github.com/ceph/%s.git' % name])
        os.chdir(vendor_src)
        run(['git', 'checkout', version])
        if cmd:
            run(cmd)
        run(['mv', vendor_module, vendor_dest])
    os.chdir(current_dir)


def clean_vendor(name):
    """
    Ensure that vendored code/dirs are removed, possibly when packaging when
    the environment flag is set to avoid vendoring.
    """
    this_dir = path.dirname(path.abspath(__file__))
    vendor_dest = path.join(this_dir, 'ceph_deploy/lib/vendor/%s' % name)
    run(['rm', '-rf', vendor_dest])


def vendorize(vendor_requirements):
    """
    This is the main entry point for vendorizing requirements. It expects
    a list of tuples that should contain the name of the library and the
    version.

    For example, a library ``foo`` with version ``0.0.1`` would look like::

        vendor_requirements = [
            ('foo', '0.0.1'),
        ]
    """

    for library in vendor_requirements:
        if len(library) == 2:
            name, version = library
            cmd = None
        elif len(library) == 3:  # a possible cmd we need to run
            name, version, cmd = library
        vendor_library(name, version, cmd)



try:
    import configparser
except ImportError:
    import ConfigParser as configparser
import logging
import os
from os import path
import re

from ceph_deploy.util.paths import gpg

logger = logging.getLogger('ceph_deploy.conf')

cd_conf_template = """
#
# ceph-deploy configuration file
#

[ceph-deploy-global]
# Overrides for some of ceph-deploy's global flags, like verbosity or cluster
# name

[ceph-deploy-install]
# Overrides for some of ceph-deploy's install flags, like version of ceph to
# install


#
# Repositories section
#

# yum repos:
# [myrepo]
# baseurl = http://gitbuilder.ceph.com/ceph-rpm-centos7-x86_64-basic/ref/hammer
# gpgurl = https://download.ceph.com/keys/autobuild.asc
# default = True
# extra-repos = cephrepo  # will install the cephrepo file too
#
# [cephrepo]
# name=ceph repo noarch packages
# baseurl=http://download.ceph.com/rpm-hammer/el6/noarch
# enabled=1
# gpgcheck=1
# type=rpm-md
# gpgkey=https://download.ceph.com/keys/release.asc

# apt repos:
# [myrepo]
# baseurl = http://gitbuilder.ceph.com/ceph-deb-trusty-x86_64-basic/ref/hammer
# gpgurl = https://download.ceph.com/keys/autobuild.asc
# default = True
# extra-repos = cephrepo  # will install the cephrepo file too
#
# [cephrepo]
# baseurl=http://download.ceph.com/debian-hammer
# gpgkey=https://download.ceph.com/keys/release.asc
""".format(gpgurl=gpg.url('release'))


def location():
    """
    Find and return the location of the ceph-deploy configuration file. If this
    file does not exist, create one in a default location.
    """
    return _locate_or_create()


def load():
    parser = Conf()
    parser.read(location())
    return parser


def _locate_or_create():
    home_config = path.expanduser('~/.cephdeploy.conf')
    # With order of importance
    locations = [
        path.join(os.getcwd(), 'cephdeploy.conf'),
        home_config,
    ]

    for location in locations:
        if path.exists(location):
            logger.debug('found configuration file at: %s' % location)
            return location
    logger.info('could not find configuration file, will create one in $HOME')
    create_stub(home_config)
    return home_config


def create_stub(_path=None):
    _path = _path or path.expanduser('~/.cephdeploy.conf')
    logger.debug('creating new configuration file: %s' % _path)
    with open(_path, 'w') as cd_conf:
        cd_conf.write(cd_conf_template)


def set_overrides(args, _conf=None):
    """
    Read the configuration file and look for ceph-deploy sections
    to set flags/defaults from the values found. This will alter the
    ``args`` object that is created by argparse.
    """
    # Get the subcommand name to avoid overwritting values from other
    # subcommands that are not going to be used
    subcommand = args.func.__name__
    command_section = 'ceph-deploy-%s' % subcommand
    conf = _conf or load()

    for section_name in conf.sections():
        if section_name in ['ceph-deploy-global', command_section]:
            override_subcommand(
                section_name,
                conf.items(section_name),
                args
            )
    return args


def override_subcommand(section_name, section_items, args):
    """
    Given a specific section in the configuration file that maps to
    a subcommand (except for the global section) read all the keys that are
    actual argument flags and slap the values for that one subcommand.

    Return the altered ``args`` object at the end.
    """
    # XXX We are not coercing here any int-like values, so if ArgParse
    # does that in the CLI we are totally non-compliant with that expectation
    # but we will try and infer a few boolean values

    # acceptable boolean states for flags
    _boolean_states = {'yes': True, 'true': True, 'on': True,
                       'no': False, 'false': False, 'off': False}

    for k, v, in section_items:
        # get the lower case value of `v`, fallback to the booleanized
        # (original) value of `v`
        try:
            normalized_value = v.lower()
        except AttributeError:
            # probably not a string object that has .lower
            normalized_value = v
        value = _boolean_states.get(normalized_value, v)
        setattr(args, k, value)
    return args


class Conf(configparser.SafeConfigParser):
    """
    Subclasses from SafeConfigParser to give a few helpers for the ceph-deploy
    configuration. Specifically, it addresses the need to work with custom
    sections that signal the usage of custom repositories.
    """

    reserved_sections = ['ceph-deploy-global', 'ceph-deploy-install']

    def get_safe(self, section, key, default=None):
        """
        Attempt to get a configuration value from a certain section
        in a ``cfg`` object but returning None if not found. Avoids the need
        to be doing try/except {ConfigParser Exceptions} every time.
        """
        try:
            return self.get(section, key)
        except (configparser.NoSectionError, configparser.NoOptionError):
            return default

    def get_repos(self):
        """
        Return all the repo sections from the config, excluding the ceph-deploy
        reserved sections.
        """
        return [
            section for section in self.sections()
            if section not in self.reserved_sections
        ]

    @property
    def has_repos(self):
        """
        boolean to reflect having (or not) any repository sections
        """
        for section in self.sections():
            if section not in self.reserved_sections:
                return True
        return False

    def get_list(self, section, key):
        """
        Assumes that the value for a given key is going to be a list
        separated by commas. It gets rid of trailing comments.
        If just one item is present it returns a list with a single item, if no
        key is found an empty list is returned.
        """
        value = self.get_safe(section, key, [])
        if value == []:
            return value

        # strip comments
        value = re.split(r'\s+#', value)[0]

        # split on commas
        value = value.split(',')

        # strip spaces
        return [x.strip() for x in value]

    def get_default_repo(self):
        """
        Go through all the repositories defined in the config file and search
        for a truthy value for the ``default`` key. If there isn't any return
        None.
        """
        for repo in self.get_repos():
            if self.get_safe(repo, 'default') and self.getboolean(repo, 'default'):
                return repo
        return False



from ceph_deploy import exc, mon
from ceph_deploy.conf.ceph import CephConf
from mock import Mock
import pytest


def make_fake_conf():
    return CephConf()

# NOTE: If at some point we re-use this helper, move it out
# and make it even more generic

def make_fake_conn(receive_returns=None):
    receive_returns = receive_returns or ([b'{}'], [], 0)
    conn = Mock()
    conn.return_value = conn
    conn.execute = conn
    conn.receive = Mock(return_value=receive_returns)
    conn.gateway.remote_exec = conn.receive
    conn.result = Mock(return_value=conn)
    return conn


class TestGetMonInitialMembers(object):

    def test_assert_if_mon_none_and_empty_True(self):
        cfg = make_fake_conf()
        with pytest.raises(exc.NeedHostError):
            mon.get_mon_initial_members(Mock(), True, cfg)

    def test_return_if_mon_none_and_empty_false(self):
        cfg = make_fake_conf()
        mon_initial_members = mon.get_mon_initial_members(Mock(), False, cfg)
        assert mon_initial_members is None

    def test_single_item_if_mon_not_none(self):
        cfg = make_fake_conf()
        cfg.add_section('global')
        cfg.set('global', 'mon initial members', 'AAAA')
        mon_initial_members = mon.get_mon_initial_members(Mock(), False, cfg)
        assert set(mon_initial_members) == set(['AAAA'])

    def test_multiple_item_if_mon_not_none(self):
        cfg = make_fake_conf()
        cfg.add_section('global')
        cfg.set('global', 'mon initial members', 'AAAA, BBBB')
        mon_initial_members = mon.get_mon_initial_members(Mock(), False, cfg)
        assert set(mon_initial_members) == set(['AAAA', 'BBBB'])


class TestCatchCommonErrors(object):

    def setup(self):
        self.logger = Mock()

    def assert_logger_message(self, logger, msg):
        calls = logger.call_args_list
        for log_call in calls:
            if msg in log_call[0][0]:
                return True
        raise AssertionError('"%s" was not found in any of %s' % (msg, calls))

    def test_warn_if_no_intial_members(self):
        fake_conn = make_fake_conn()
        cfg = make_fake_conf()
        mon.catch_mon_errors(fake_conn, self.logger, 'host', cfg, Mock())
        expected_msg = 'is not defined in `mon initial members`'
        self.assert_logger_message(self.logger.warning, expected_msg)

    def test_warn_if_host_not_in_intial_members(self):
        fake_conn = make_fake_conn()
        cfg = make_fake_conf()
        cfg.add_section('global')
        cfg.set('global', 'mon initial members', 'AAAA')
        mon.catch_mon_errors(fake_conn, self.logger, 'host', cfg, Mock())
        expected_msg = 'is not defined in `mon initial members`'
        self.assert_logger_message(self.logger.warning, expected_msg)

    def test_warn_if_not_mon_in_monmap(self):
        fake_conn = make_fake_conn()
        cfg = make_fake_conf()
        mon.catch_mon_errors(fake_conn, self.logger, 'host', cfg, Mock())
        expected_msg = 'does not exist in monmap'
        self.assert_logger_message(self.logger.warning, expected_msg)

    def test_warn_if_not_public_addr_and_not_public_netw(self):
        fake_conn = make_fake_conn()
        cfg = make_fake_conf()
        cfg.add_section('global')
        mon.catch_mon_errors(fake_conn, self.logger, 'host', cfg, Mock())
        expected_msg = 'neither `public_addr` nor `public_network`'
        self.assert_logger_message(self.logger.warning, expected_msg)



import logging

from ceph_deploy.util import templates
from ceph_deploy.lib import remoto
from ceph_deploy.hosts.common import map_components

LOG = logging.getLogger(__name__)

NON_SPLIT_PACKAGES = ['ceph-osd', 'ceph-mon', 'ceph-mds']


def install(distro, version_kind, version, adjust_repos, **kw):
    packages = map_components(
        NON_SPLIT_PACKAGES,
        kw.get('components', [])
    )

    distro.packager.clean()
    if packages:
        distro.packager.install(packages)


def mirror_install(distro, repo_url, gpg_url, adjust_repos, **kw):
    packages = map_components(
        NON_SPLIT_PACKAGES,
        kw.get('components', [])
    )
    repo_url = repo_url.strip('/')  # Remove trailing slashes
    gpg_url_path = gpg_url.split('file://')[-1]  # Remove file if present
    gpgcheck = kw.pop('gpgcheck', 1)

    if adjust_repos:
        remoto.process.run(
            distro.conn,
            [
                'rpm',
                '--import',
                gpg_url_path,
            ]
        )

        ceph_repo_content = templates.zypper_repo.format(
            repo_url=repo_url,
            gpg_url=gpg_url,
            gpgcheck=gpgcheck,
        )
        distro.conn.remote_module.write_file(
            '/etc/zypp/repos.d/ceph.repo',
            ceph_repo_content.encode('utf-8'))
        distro.packager.clean()

    if packages:
        distro.packager.install(packages)


def repo_install(distro, reponame, baseurl, gpgkey, **kw):
    packages = map_components(
        NON_SPLIT_PACKAGES,
        kw.pop('components', [])
    )
    # Get some defaults
    name = kw.get('name', '%s repo' % reponame)
    enabled = kw.get('enabled', 1)
    gpgcheck = kw.get('gpgcheck', 1)
    install_ceph = kw.pop('install_ceph', False)
    proxy = kw.get('proxy')
    _type = 'repo-md'
    baseurl = baseurl.strip('/')  # Remove trailing slashes

    if gpgkey:
        remoto.process.run(
            distro.conn,
            [
                'rpm',
                '--import',
                gpgkey,
            ]
        )

    repo_content = templates.custom_repo(
        reponame=reponame,
        name = name,
        baseurl = baseurl,
        enabled = enabled,
        gpgcheck = gpgcheck,
        _type = _type,
        gpgkey = gpgkey,
        proxy = proxy,
    )

    distro.conn.remote_module.write_file(
        '/etc/zypp/repos.d/%s' % (reponame),
        repo_content.encode('utf-8')
    )

    # Some custom repos do not need to install ceph
    if install_ceph and packages:
        distro.packager.install(packages)



import subprocess

import pytest
from mock import Mock, patch

from ceph_deploy.cli import _main as main
from ceph_deploy.tests.directory import directory
from ceph_deploy.tests.util import assert_too_few_arguments


#TODO: This test does check that things fail if the .conf file is missing
def test_bad_no_conf(tmpdir, cli):
    with pytest.raises(cli.Failed) as err:
        with cli(
            args=['ceph-deploy', 'mon'],
            stderr=subprocess.PIPE,
            ) as p:
            result = p.stderr.read().decode('utf-8')
    assert 'usage: ceph-deploy' in result
    assert_too_few_arguments(result)
    assert err.value.status == 2


def make_fake_connection(platform_information=None):
    get_connection = Mock()
    get_connection.return_value = get_connection
    get_connection.remote_module.platform_information = Mock(
        return_value=platform_information)
    return get_connection


def test_new(tmpdir, capsys):
    with tmpdir.join('ceph.conf').open('w') as f:
        f.write("""\
[global]
fsid = 6ede5564-3cf1-44b5-aa96-1c77b0c3e1d0
mon initial members = host1
""")

    fake_ip_addresses = lambda x: ['10.0.0.1']
    try:
        with patch('ceph_deploy.new.net.ip_addresses', fake_ip_addresses):
            with patch('ceph_deploy.new.net.get_nonlocal_ip', lambda x: '10.0.0.1'):
                with patch('ceph_deploy.new.arg_validators.Hostname', lambda: lambda x: x):
                    with patch('ceph_deploy.new.hosts'):
                        with directory(str(tmpdir)):
                            main(['-v', 'new', '--no-ssh-copykey', 'host1'])
    except SystemExit as e:
        raise AssertionError('Unexpected exit: %s', e)
    out, err = capsys.readouterr()
    err = err.lower()
    assert 'creating new cluster named ceph' in err
    assert 'monitor host1 at 10.0.0.1' in err
    assert 'resolving host host1' in err
    assert "monitor initial members are ['host1']" in err
    assert "monitor addrs are ['10.0.0.1']" in err



import re
import uuid

from mock import patch

from ceph_deploy import conf
from ceph_deploy.cli import _main as main
from ceph_deploy.tests.directory import directory


def test_write_global_conf_section(tmpdir):
    fake_ip_addresses = lambda x: ['10.0.0.1']

    with patch('ceph_deploy.new.hosts'):
        with patch('ceph_deploy.new.net.ip_addresses', fake_ip_addresses):
            with patch('ceph_deploy.new.net.get_nonlocal_ip', lambda x: '10.0.0.1'):
                with patch('ceph_deploy.new.arg_validators.Hostname', lambda: lambda x: x):
                    with directory(str(tmpdir)):
                        main(args=['new', 'host1'])
    with tmpdir.join('ceph.conf').open() as f:
        cfg = conf.ceph.parse(f)
    assert cfg.sections() == ['global']


def pytest_funcarg__newcfg(request):
    tmpdir = request.getfuncargvalue('tmpdir')
    fake_ip_addresses = lambda x: ['10.0.0.1']

    def new(*args):
        with patch('ceph_deploy.new.net.ip_addresses', fake_ip_addresses):
            with patch('ceph_deploy.new.hosts'):
                with patch('ceph_deploy.new.net.get_nonlocal_ip', lambda x: '10.0.0.1'):
                    with patch('ceph_deploy.new.arg_validators.Hostname', lambda: lambda x: x):
                        with directory(str(tmpdir)):
                            main(args=['new'] + list(args))
                            with tmpdir.join('ceph.conf').open() as f:
                                cfg = conf.ceph.parse(f)
                            return cfg
    return new


def test_uuid(newcfg):
    cfg = newcfg('host1')
    fsid = cfg.get('global', 'fsid')
    # make sure it's a valid uuid
    uuid.UUID(hex=fsid)
    # make sure it looks pretty, too
    UUID_RE = re.compile(
        r'^[0-9a-f]{8}-'
        + r'[0-9a-f]{4}-'
        # constant 4 here, we want to enforce randomness and not leak
        # MACs or time
        + r'4[0-9a-f]{3}-'
        + r'[0-9a-f]{4}-'
        + r'[0-9a-f]{12}$',
        )
    assert UUID_RE.match(fsid)


def test_mons(newcfg):
    cfg = newcfg('node01', 'node07', 'node34')
    mon_initial_members = cfg.get('global', 'mon_initial_members')
    assert mon_initial_members == 'node01, node07, node34'


def test_defaults(newcfg):
    cfg = newcfg('host1')
    assert cfg.get('global', 'auth cluster required') == 'cephx'
    assert cfg.get('global', 'auth service required') == 'cephx'
    assert cfg.get('global', 'auth client required') == 'cephx'



"""
This module is meant for vendorizing Python libraries. Most libraries will need
to have some ``sys.path`` alterations done unless they are doing relative
imports.

Do **not** add anything to this module that does not represent a vendorized
library.

Vendored libraries should go into the ``vendor`` directory and imported from
there. This is so we allow libraries that are installed normally to be imported
if the vendored module is not available.

The import dance here is done so that all other imports throught ceph-deploy
are kept the same regardless of where the module comes from.

The expected way to import execnet would look like this::

    from remoto.lib import execnet

"""
import sys
import os
this_dir = os.path.abspath(os.path.dirname(__file__))
vendor_dir = os.path.join(this_dir, 'vendor')

try:
    # vendored
    if vendor_dir not in sys.path:
        sys.path.insert(0, vendor_dir)
    import execnet
except ImportError as err:
    # normally installed
    import execnet  # noqa






def generate_ips(start_ip, end_ip):
    start = list(map(int, start_ip.split(".")))
    end = list(map(int, end_ip.split(".")))
    temp = start
    ip_range = []

    ip_range.append(start_ip)
    while temp != end:
        start[3] += 1
        for i in (3, 2, 1):
            if temp[i] == 256:
                temp[i] = 0
                temp[i-1] += 1
        ip_range.append(".".join(map(str, temp)))

    return ip_range


class Empty(object):
    """
    A bare class, with explicit behavior for key/value items to be set at
    instantiation.
    """
    def __init__(self, **kw):
        for k, v in kw.items():
            setattr(self, k, v)


def assert_too_few_arguments(err):
    assert ("error: too few arguments" in err or
            "error: the following argument" in err)



try:
    from cStringIO import StringIO
except ImportError:
    from io import StringIO
from ceph_deploy import conf


def test_simple():
    f = StringIO("""\
[foo]
bar = baz
""")
    cfg = conf.ceph.parse(f)
    assert cfg.get('foo', 'bar') == 'baz'


def test_indent_space():
    f = StringIO("""\
[foo]
        bar = baz
""")
    cfg = conf.ceph.parse(f)
    assert cfg.get('foo', 'bar') == 'baz'


def test_indent_tab():
    f = StringIO("""\
[foo]
\tbar = baz
""")
    cfg = conf.ceph.parse(f)
    assert cfg.get('foo', 'bar') == 'baz'


def test_words_underscore():
    f = StringIO("""\
[foo]
bar_thud = baz
""")
    cfg = conf.ceph.parse(f)
    assert cfg.get('foo', 'bar_thud') == 'baz'
    assert cfg.get('foo', 'bar thud') == 'baz'


def test_words_space():
    f = StringIO("""\
[foo]
bar thud = baz
""")
    cfg = conf.ceph.parse(f)
    assert cfg.get('foo', 'bar_thud') == 'baz'
    assert cfg.get('foo', 'bar thud') == 'baz'


def test_words_many():
    f = StringIO("""\
[foo]
bar__ thud   quux = baz
""")
    cfg = conf.ceph.parse(f)
    assert cfg.get('foo', 'bar_thud_quux') == 'baz'
    assert cfg.get('foo', 'bar thud quux') == 'baz'


def test_write_words_underscore():
    cfg = conf.ceph.CephConf()
    cfg.add_section('foo')
    cfg.set('foo', 'bar thud quux', 'baz')
    f = StringIO()
    cfg.write(f)
    f.seek(0)
    assert f.readlines() == ['[foo]\n', 'bar_thud_quux = baz\n','\n']


def test_section_repeat():
    f = StringIO("""\
[foo]
bar = bez
thud = quux

[foo]
bar = baz
""")
    cfg = conf.ceph.parse(f)
    assert cfg.get('foo', 'bar') == 'baz'
    assert cfg.get('foo', 'thud') == 'quux'



import os
import logging

from ceph_deploy import hosts
from ceph_deploy.cliutil import priority


LOG = logging.getLogger(__name__)


def install_repo(distro, args, cd_conf, rlogger):
    if args.repo_name in cd_conf.get_repos():
        LOG.info('will use repository %s from ceph-deploy config', args.repo_name)
        options = dict(cd_conf.items(args.repo_name))
        extra_repos = cd_conf.get_list(args.repo_name, 'extra-repos')
        try:
            repo_url = options.pop('baseurl')
            gpg_url = options.pop('gpgkey', None)
        except KeyError as err:
            raise RuntimeError(
                'missing required key: %s in config section: %s' % (err, args.repo_name)
            )
    else:
        repo_url = os.environ.get('CEPH_DEPLOY_REPO_URL') or args.repo_url
        gpg_url = os.environ.get('CEPH_DEPLOY_GPG_URL') or args.gpg_url
        extra_repos = []

    repo_url = repo_url.strip('/')  # Remove trailing slashes
    distro.packager.add_repo(
        args.repo_name,
        repo_url,
        gpg_url=gpg_url
    )

    for xrepo in extra_repos:
        rlogger.info('adding extra repo: %s' % xrepo)
        options = dict(cd_conf.items(xrepo))
        try:
            repo_url = options.pop('baseurl')
            gpg_url = options.pop('gpgkey', None)
        except KeyError as err:
            raise RuntimeError(
                'missing required key: %s in config section: %s' % (err, xrepo)
            )
        distro.packager.add_repo(
            args.repo_name,
            repo_url,
            gpg_url=gpg_url
        )


def repo(args):
    cd_conf = getattr(args, 'cd_conf', None)

    for hostname in args.host:
        LOG.debug('Detecting platform for host %s ...', hostname)
        distro = hosts.get(
            hostname,
            username=args.username
        )
        rlogger = logging.getLogger(hostname)

        LOG.info(
            'Distro info: %s %s %s',
            distro.name,
            distro.release,
            distro.codename
        )

        if args.remove:
            distro.packager.remove_repo(args.repo_name)
        else:
            install_repo(distro, args, cd_conf, rlogger)


@priority(70)
def make(parser):
    """
    Repo definition management
    """

    parser.add_argument(
        'repo_name',
        metavar='REPO-NAME',
        help='Name of repo to manage.  Can match an entry in cephdeploy.conf'
    )

    parser.add_argument(
        '--repo-url',
        help='a repo URL that mirrors/contains Ceph packages'
    )

    parser.add_argument(
        '--gpg-url',
        help='a GPG key URL to be used with custom repos'
    )

    parser.add_argument(
        '--remove', '--delete',
        action='store_true',
        help='remove repo definition on remote host'
    )

    parser.add_argument(
        'host',
        metavar='HOST',
        nargs='+',
        help='host(s) to install on'
    )

    parser.set_defaults(
        func=repo
    )



import pytest
from mock import patch

from ceph_deploy.cli import get_parser
from ceph_deploy.tests.fakes import fake_arg_val_hostname
from ceph_deploy.tests.util import assert_too_few_arguments

@patch('ceph_deploy.util.arg_validators.Hostname.__call__', fake_arg_val_hostname)
class TestParserNew(object):

    def setup(self):
        self.parser = get_parser()

    def test_help(self, capsys):
        with pytest.raises(SystemExit):
            self.parser.parse_args('new --help'.split())
        out, err = capsys.readouterr()
        assert 'usage: ceph-deploy new' in out
        assert 'positional arguments:' in out
        assert 'optional arguments:' in out

    def test_new_copykey_true_by_default(self):
        args = self.parser.parse_args('new host1'.split())
        assert args.ssh_copykey

    def test_new_copykey_false(self):
        args = self.parser.parse_args('new --no-ssh-copykey host1'.split())
        assert not args.ssh_copykey

    def test_new_fsid_none_by_default(self):
        args = self.parser.parse_args('new host1'.split())
        assert args.fsid is None

    def test_new_fsid_custom_fsid(self):
        args = self.parser.parse_args('new --fsid bc50d015-65c9-457a-bfed-e37b92756527 host1'.split())
        assert args.fsid == 'bc50d015-65c9-457a-bfed-e37b92756527'

    @pytest.mark.skipif(reason="no UUID validation yet")
    def test_new_fsid_custom_fsid_bad(self, capsys):
        with pytest.raises(SystemExit):
            self.parser.parse_args('new --fsid bc50d015-65c9-457a-bfed-e37'.split())
        out, err = capsys.readouterr()
        #TODO check for correct error string in err

    def test_new_networks_none_by_default(self):
        args = self.parser.parse_args('new host1'.split())
        assert args.public_network is None
        assert args.cluster_network is None

    def test_new_public_network_custom(self):
        args = self.parser.parse_args('new --public-network 10.10.0.0/16 host1'.split())
        assert args.public_network == "10.10.0.0/16"

    def test_new_cluster_network_custom(self):
        args = self.parser.parse_args('new --cluster-network 10.10.0.0/16 host1'.split())
        assert args.cluster_network == "10.10.0.0/16"

    def test_new_public_network_custom_bad(self, capsys):
        with pytest.raises(SystemExit):
            self.parser.parse_args('new --public-network 10.10.0.0'.split())
        out, err = capsys.readouterr()
        assert "error: subnet must" in err

    def test_new_cluster_network_custom_bad(self, capsys):
        with pytest.raises(SystemExit):
            self.parser.parse_args('new --cluster-network 10.10.0.0'.split())
        out, err = capsys.readouterr()
        assert "error: subnet must" in err

    def test_new_mon_required(self, capsys):
        with pytest.raises(SystemExit):
            self.parser.parse_args('new'.split())
        out, err = capsys.readouterr()
        assert_too_few_arguments(err)

    def test_new_one_mon(self):
        hostnames = ['test1']
        args = self.parser.parse_args(['new'] + hostnames)
        assert args.mon == hostnames

    def test_new_multiple_mons(self):
        hostnames = ['test1', 'test2', 'test3']
        args = self.parser.parse_args(['new'] + hostnames)
        assert frozenset(args.mon) == frozenset(hostnames)



import pytest

import ceph_deploy
from ceph_deploy.cli import get_parser
from ceph_deploy.tests.util import assert_too_few_arguments


SUBCMDS_WITH_ARGS = [
    'new', 'install', 'rgw', 'mds', 'mon', 'gatherkeys', 'disk', 'osd',
    'admin', 'config', 'uninstall', 'purgedata', 'purge', 'pkg', 'calamari'
]
SUBCMDS_WITHOUT_ARGS = ['forgetkeys']


class TestParserMain(object):

    def setup(self):
        self.parser = get_parser()

    def test_verbose_true(self):
        args = self.parser.parse_args('--verbose forgetkeys'.split())
        assert args.verbose

    def test_verbose_default_is_false(self):
        args = self.parser.parse_args('forgetkeys'.split())
        assert not args.verbose

    def test_quiet_true(self):
        args = self.parser.parse_args('--quiet forgetkeys'.split())
        assert args.quiet

    def test_quiet_default_is_false(self):
        args = self.parser.parse_args('forgetkeys'.split())
        assert not args.quiet

    def test_verbose_quiet_are_mutually_exclusive(self, capsys):
        with pytest.raises(SystemExit):
            self.parser.parse_args('--verbose --quiet forgetkeys'.split())
        out, err = capsys.readouterr()
        assert 'not allowed with argument' in err

    def test_version(self, capsys):
        with pytest.raises(SystemExit):
            self.parser.parse_args('--version'.split())
        out, err = capsys.readouterr()
        assert ceph_deploy.__version__ in (out.strip(), err.strip())

    def test_custom_username(self):
        args = self.parser.parse_args('--username trhoden forgetkeys'.split())
        assert args.username == 'trhoden'

    def test_default_username_is_none(self):
        args = self.parser.parse_args('forgetkeys'.split())
        assert args.username is None

    def test_overwrite_conf_default_false(self):
        args = self.parser.parse_args('forgetkeys'.split())
        assert not args.overwrite_conf

    def test_overwrite_conf_true(self):
        args = self.parser.parse_args('--overwrite-conf forgetkeys'.split())
        assert args.overwrite_conf

    def test_default_cluster_name(self):
        args = self.parser.parse_args('forgetkeys'.split())
        assert args.cluster == 'ceph'

    def test_custom_cluster_name(self):
        args = self.parser.parse_args('--cluster myhugecluster forgetkeys'.split())
        assert args.cluster == 'myhugecluster'

    def test_custom_cluster_name_bad(self, capsys):
        with pytest.raises(SystemExit):
            self.parser.parse_args('--cluster=/evil-this-should-not-be-created'.split())
        out, err = capsys.readouterr()
        assert ('--cluster: argument must start with a letter and contain only '
                'letters and numbers') in err

    def test_default_ceph_conf_is_none(self):
        args = self.parser.parse_args('forgetkeys'.split())
        assert args.ceph_conf is None

    def test_custom_ceph_conf(self):
        args = self.parser.parse_args('--ceph-conf /tmp/ceph.conf forgetkeys'.split())
        assert args.ceph_conf == '/tmp/ceph.conf'

    @pytest.mark.parametrize('cmd', SUBCMDS_WITH_ARGS)
    def test_valid_subcommands_with_args(self, cmd, capsys):
        with pytest.raises(SystemExit):
            self.parser.parse_args(['%s' % cmd])
        out, err = capsys.readouterr()
        assert_too_few_arguments(err)
        assert 'invalid choice' not in err

    @pytest.mark.parametrize('cmd', SUBCMDS_WITHOUT_ARGS)
    def test_valid_subcommands_without_args(self, cmd, capsys):
        self.parser.parse_args(['%s' % cmd])

    def test_invalid_subcommand(self, capsys):
        with pytest.raises(SystemExit):
            self.parser.parse_args('bork'.split())
        out, err = capsys.readouterr()
        assert 'invalid choice' in err

    def test_help(self, capsys):
        with pytest.raises(SystemExit):
            self.parser.parse_args('--help'.split())
        out, err = capsys.readouterr()
        assert 'usage: ceph-deploy' in out
        assert 'optional arguments:' in out
        assert 'commands:' in out





class NormalizedVersion(object):
    """
    A class to provide a clean interface for setting/retrieving distinct
    version parts divided into major, minor, and patch (following convnetions
    from semver (see http://semver.org/)

    Since a lot of times version parts need to be compared, it provides for
    `int` representations of their string counterparts, with some sanitization
    processing.

    Defaults to '0' or 0 (int) values when values are not set or parsing fails.
    """

    def __init__(self, raw_version):
        self.raw_version = raw_version.strip()
        self.major = '0'
        self.minor = '0'
        self.patch = '0'
        self.garbage = ''
        self.int_major = 0
        self.int_minor = 0
        self.int_patch = 0
        self._version_map = {}
        self._set_versions()

    def _set_int_versions(self):
        version_map = dict(
            major=self.major,
            minor=self.minor,
            patch=self.patch,
            garbage=self.garbage)

        # safe int versions that remove non-numerical chars
        # for example 'rc1' in a version like '1-rc1
        for name, value in version_map.items():
            if '-' in value:  # get rid of garbage like -dev1 or -rc1
                value = value.split('-')[0]
            value = float(''.join(c for c in value if c.isdigit()) or 0)
            int_name = "int_%s" % name
            setattr(self, int_name, value)

    def _set_versions(self):
        split_version = (self.raw_version.split('.') + ["0"]*4)[:4]
        self.major, self.minor, self.patch, self.garbage = split_version
        self._set_int_versions()
